"""
Common code for all runtimes.

Desktop runtimes provide a way to load a page as desktop application. To do
so, we may cache or link to an existing runtime (like Firefox) or even
keep a local copy of a runtime (like NW.js). Some runtimes (e.g. Firefox)
needs a whole directory structure as its app definition. Others need just
one manifest file (NW.js) and perhaps some icons. Others need nothing (Chrome).
In all cases we improve the app experience by making use of a custom named
executable to control task bar grouping, and on OS X we build an actual
application (xx.app directory).

It is assumed that desktop runtimes are backward compatible. This is a
reasonable assumption since we use only the web stuff, which browsers
generally keep working.

We also don't make a point of always having the latest version, because
some runtimes release almost every week. Having the user confirm a
download such often is way too much a burden, and auto-update too
complex / error-prone. These updates are mostly for security reasons,
which is generally less an issue for us because we only connect them
to known sources which are on localhost for desktop apps anyway.

Therefore, the webruntime module has a hardcoded minimal version for runtimes
where this makes sense, which is configurable by the user in cases where its
needed.

For deskop runtimes we have the following important attributes:

* icon: the application icon for the app. This will usually come from the
  main widget's icon property (as a string), and is converted to an Icon
  object here, so that each runtime can export the required .ico, .icns or .png
  files.
* title: the title to display on the apps title bar. This will usually come from
  the main widget's title property.
* exe_name: the name of the executable of the runtime, chosing this helps
  find the process in the task manager, but more importantly, avoids task
  grouping, or helps wanted grouping.
* id: a unique application id, generated and used internally to creat unique
  temporary app dirs and application manifests.

"""

import os.path as op
import os
import sys
import time
import signal
import atexit
import shutil
import threading
import subprocess

from . import logger
from .util.icon import Icon
from ._manage import RUNTIME_DIR
from ._manage import init_dirs, clean_dirs, lock_runtime_dir, versionstring
from ._defaulticon import default_icon


INFO_PLIST = """
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" NONL
"http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>CFBundleIconFile</key>
    <string>app.icns</string>
    <key>CFBundleDevelopmentRegion</key>
    <string>English</string>
    <key>CFBundleExecutable</key>
    <string>{exe}</string>
    <key>CFBundleName</key>
    <string>{name}</string>
</dict>
</plist>
""".lstrip().replace('    ', '\t').replace('NONL\n', '')


class BaseRuntime:
    """ Base class for all runtimes.
    """
    
    def __init__(self, **kwargs):
        
        # nomnom, we eat all kwargs, because different runtimes use
        # different kwargs, and we want it to be easy to switch runtimes
        self._leftover_kwargs = kwargs
        
        assert self.get_name()
        
        # Close the runtime when Python closes cleanly
        atexit.register(self.close)
        
        # Increase chance of closing runtime when Python is forced to stop
        signal.signal(signal.SIGTERM, self.close)
        signal.signal(signal.SIGINT, self.close)
        
        self._exe = None
        self._version = None
        self._proc = None
        self._streamreader = None
        
        # Tidy up, but don't make us wait for it, also give main thread
        # time to e.g. continue incomplete downloads (e.g. for NW.js runtime)
        init_dirs()
        t = threading.Thread(target=lambda: time.sleep(4) or clean_dirs())
        t.setDaemon(True)
        t.start()  # tidy up
    
    def get_install_instuctions(self):
        """ Get instructions on how a runtime can be installed. Used internally
        to show useful dialogs.
        """
        return self._get_install_instuctions().strip()
    
    def get_name(self):
        """ Get the name of the runtime.
        """
        return self._get_name()
    
    def get_exe(self):
        """ Get the executable corresponding to the runtime. This is usually
        the path to an executable file, but it can also be a command, or even
        a stub. Can be None if the runtime is not (yet) installed.
        """
        if not self._exe:
            self._exe = self._get_exe()
        return self._exe
    
    def get_version(self):
        """ Get the version of the available runtime (as a string). Can be None
        if the version cannot be retrieved (e.g. for Edge and IE), or if the
        runtime is not available on this system.
        """
        if not self._version:
            self._version = self._get_version()
        return self._version
    
    def is_available(self):
        """ Get whether this runtime appears to be available on this machine.
        """
        return self._is_available()
    
    def launch_tab(self, url):
        """ Launch the given url in a new browser tab. Only works for runtimes
        that are browsers (e.g. not NW).
        """
        if not self.is_available():
            t = 'Cannot launch tab, because %s runtime is not available'
            raise RuntimeError(t % self.get_name())
        self._launch_tab(url)
        logger.info('launched in %s tab: %s' % (self.get_name(), url))
    
    def launch_app(self, url):
        """ Launch the given url as a desktop application. Only works for
        runtimes that derive from DeskopRuntime (e.g. not Edge). Apps launched
        this way can usually be terminated using the ``close()`` method.
        """
        if not self.is_available():
            t = 'Cannot launch app, because %s runtime is not available'
            raise RuntimeError(t % self.get_name())
        self._launch_app(url)
        logger.info('launched as %s app: %s' % (self.get_name(), url))
    
    def close(self):
        """ Close the runtime, or kill it if the process does not
        respond. Note that closing only works for runtimes launched as
        an app (using ``launch_app()``).
        """
        if self._proc is None:
            return
        # Terminate, wait for a bit, kill
        self._proc.we_closed_it = True
        if self._proc.poll() is None:
            if self._proc.stdin:  # pragma: no cover
                self._proc.stdin.close()
            self._proc.terminate()
            timeout = time.time() + 0.2
            while time.time() < timeout:
                time.sleep(0.02)
                if self._proc.poll() is not None:
                    break
            else:  # pragma: no cover
                self._proc.kill()
        # Discart process
        self._proc = None
    
    ## Utilities that this class provides for subclasses
    
    def _start_subprocess(self, cmd, shell=False, **env):
        """ Start subclasses, store handle, and launch a thread to read
        stdout for the process. Intended for web runtimes that are "bound"
        to this process.
        """
        if self._proc:
            t = 'Cannot launch %s app twice with same runtime instance.'
            raise RuntimeError(t % self.get_name())
        self._proc = self._spawn_subprocess(cmd, shell, **env)
        self._streamreader = StreamReader(self._proc)
        self._streamreader.start()
    
    def _spawn_subprocess(self, cmd, shell=False, **env):
        """ Spawn a subprocess and return process handle. Intended for
        web runtimes that are "spawned", like browsers.
        """
        environ = os.environ.copy()
        environ.update(env)
        try:
            return subprocess.Popen(cmd, env=environ, shell=shell,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT)
        except OSError as err:  # pragma: no cover
            raise RuntimeError('Could not start runtime with command %r:\n%s' %
                               (cmd[0], str(err)))
    
    ## Methods to implement in subclasses
    
    def _get_install_instuctions(self):
        """ Subclasses can let know how the user must install the runtime.
        Can return None to indicate that it cannot be installed (like the
        "default browser").
        """
        return "No info on how to install %s" % self.get_name()
    
    def _get_name(self):
        """ Just make this return a string name.
        """
        raise NotImplementedError()
    
    def _get_exe(self):
        """ String executable name, preferably an absolute path. Should return
        something if the runtime appears to be available.
        """
        raise NotImplementedError()
    
    def _is_available(self):
        """ Alias of get_exe(), unless overloaded.
        """
        return self.get_exe()
    
    def _get_version(self):
        """ String version of the available runtime, can return None
        if the version cannot be retrieved.
        """
        raise NotImplementedError()
    
    def _launch_tab(self, url):
        """ Function to implement launching the url in a new tab.
        """
        raise NotImplementedError()
    
    def _launch_app(self, url):
        """ Function to implement launching the url as a desktop app.
        """
        raise NotImplementedError()


class DesktopRuntime(BaseRuntime):
    """ A base class for runtimes that launch a desktop-like app.

    Arguments:
        title (str): Text shown in the title bar.
        icon (str | Icon): Icon instance or path to an icon file (png or ico).
            The icon will automatically be converted to png/ico/icns,
            depending on what's needed by the runtime and platform.
        size (tuple of ints): The size (in pixels) of the window.
        pos (tuple of ints): The position (in pixels) of the window.
        windowmode (str): the initial window mode, e.g. 'normal', 'maximized',
            'fullscreen', 'kiosk'. Note that not all modes are supported by all
            runtimes.
    """

    def __init__(self, icon=None, title=None,
                 size=None, pos=None, windowmode=None, **kwargs):
        
        self._icon = iconize(icon or None)
        assert isinstance(self._icon, Icon)
        
        self._title = title or '%s webruntime' % self.get_name()
        assert isinstance(self._title, str)
        
        self._size = size or (640, 480)
        assert isinstance(self._size, tuple) and len(self._size) == 2
        
        self._pos = pos
        assert self._pos is None or (isinstance(self._pos, tuple) and
                                     len(self._pos) == 2)
        
        self._windowmode = windowmode or 'normal'
        assert isinstance(self._windowmode, str)
        assert self._windowmode in ('normal', 'maximized', 'fullscreen', 'kiosk')
        
        self._system_version = None
        
        super().__init__(**kwargs)
    
    def _is_available(self):
        # cached, _ = self.get_cached_version()
        # -> only claim to be available if the app is actually installed
        cached = None
        system, _ = self.get_system_version()
        return bool(cached or system)
    
    def get_runtime_dir(self):
        """ Get the directory of the (up-to-date) local version of the runtime.
        If necessary, the runtime is cached/installed.
        """
        
        # Get info on cached version
        cur_version, path = self.get_cached_version()
        sys_version, sys_path = self.get_system_version()
        
        # Do we need an install?
        install_action = None
        if not (cur_version or sys_version):
            raise RuntimeError('Dont have cached version of runtime %s nor can '
                               'install it.' % self.get_name())
        elif not cur_version:
            install_action = 'install'
        elif not sys_version:
            # No specific version required, e.g. because can assume that we
            # have an up-to-date version.
            pass
        elif versionstring(cur_version) < versionstring(sys_version):
            install_action = 'update'
        
        # Install if necessary and update version
        if install_action:
            logger.info('Performing %s of runtime %s' %
                        (install_action, self.get_name()))
            path = op.join(RUNTIME_DIR, self.get_name() + '_' +sys_version)
            self._install_runtime(sys_path, path)
        
        assert os.path.isdir(path)
        lock_runtime_dir(path)
        return path
    
    def get_system_version(self):
        """ Get (version, path) for the version of the runtime available
        on the system. Can be an application, but also an archive or installer.
        This information is used internally to install or update a runtime.
        """
        if not self._system_version:
            self._system_version = self._get_system_version()
            assert len(self._system_version) == 2
        return self._system_version
    
    def get_cached_version(self):
        """ Get (version, path) for the (highest) version of this runtime that
        we currently have locally installed.
        """
        versions = []
        for dname in os.listdir(RUNTIME_DIR):
            dirname = op.join(RUNTIME_DIR, dname)
            if op.isdir(dirname) and dname.startswith(self.get_name() + '_'):
                versions.append((dname.split('_')[-1], dirname))
        versions.sort(key=lambda x: versionstring(x[0]))
        if versions:
            return versions[-1]
        return None, None
    
    def get_frozen_version(self):
        """ Get (version, path) for the (highest) version of this runtime that
        is shipped along with the executable.
        """
        # todo: put this to use
        basedir = os.path.dirname(sys.executable)
        versions = []
        for dname in os.listdir(basedir):
            dirname = op.join(basedir, dname)
            if op.isdir(dirname) and dname.startswith(self.get_name() + '_'):
                versions.append((dname.split('_')[-1], dirname))
        versions.sort(key=lambda x: versionstring(x[0]))
        if versions:
            return versions[-1]
        return None, None
    
    def _get_app_exe(self, runtime_exe, app_path):
        """ Get the executable to run our app. This should take care
        that the runtime process shows up in the task manager with the
        correct exe_name.

        * runtime_exe: the location of the runtime executable (can be a symlink)
        * app_path: the location of the temp app (the app.json or whatever)

        """
        # Define process name, so that our window is not grouped with
        # Firefox, NW.js or whatever, and has a more meaningful name in the
        # task manager. Using sys.executable also works well when frozen.
        exe_name = op.basename(sys.executable)
        if sys.platform.startswith("win"):
            exe_name, ext = op.splitext(exe_name)
            exe_name = exe_name + '-ui' + ext
        else:
            exe_name = exe_name + '-ui'
        
        assert runtime_exe.startswith(RUNTIME_DIR)
        
        if sys.platform.startswith('darwin'):
            # OSX: create an app, the name of the exe does not matter
            # much but the name to give the application does. We set
            # the latter to the title, because title and process name
            # seem the same thing in osx.
            app_exe = op.join(app_path, exe_name + '.app')
            # todo: double check to make sure if title makes the most sense here
            self._osx_create_app(op.realpath(runtime_exe), app_exe, self._title)
            app_exe += '/Contents/MacOS/' + exe_name
        else:
            
            if sys.platform.startswith('win'):
                # Windows: make a copy of the executable
                app_exe = op.join(op.dirname(runtime_exe), exe_name)
                if not op.isfile(app_exe):
                    shutil.copy2(runtime_exe, app_exe)
            else:
                # Linux, create a symlink
                app_exe = op.join(app_path, exe_name)
                if not op.isfile(app_exe):
                    os.symlink(op.realpath(runtime_exe), app_exe)
        
        return app_exe

    def _osx_create_app(self, exe, dst_dir, title):
        """ Create osx app

        * exe: path to executable of runtime (not the symlink)
        * dst_dir: path of the .app directory to create.
        * title: the title of the window *and* the process name
        """

        # Get original app to copy it from
        exe_name_src = op.basename(exe)
        exe_name_dst = op.basename(dst_dir).split('.')[0]
        if 'Contents/MacOS' not in exe:
            raise NotImplementedError('Can only create OS X app from existing app')
        src_dir = op.dirname(op.dirname(op.dirname(exe)))
        if not src_dir.endswith('.app'):
            raise TypeError('The original OS X application must end in .app.')

        # Clear destination
        if op.isdir(dst_dir):
            shutil.rmtree(dst_dir)
        os.mkdir(dst_dir)

        # Make dir structure
        os.mkdir(op.join(dst_dir, 'Contents'))
        os.mkdir(op.join(dst_dir, 'Contents', 'MacOS'))

        # Make a link for all the files
        for dirpath, dirnames, filenames in os.walk(src_dir):
            relpath = op.relpath(dirpath, src_dir)
            if relpath.startswith(('Contents/MacOS', 'Contents/Resources',
                                   'Contents/Versions')):
                if not op.isdir(op.join(dst_dir, relpath)):
                    os.mkdir(op.join(dst_dir, relpath))
                for fname in filenames:
                    os.link(op.join(src_dir, relpath, fname),
                            op.join(dst_dir, relpath, fname))
        # Make runtime exe
        os.link(  # or shutil.copy2(
            op.realpath(op.join(src_dir, 'Contents', 'MacOS', exe_name_src)),
            op.join(dst_dir, 'Contents', 'MacOS', exe_name_dst))
        # Make info.plist
        info = INFO_PLIST.format(name=title, exe=exe_name_dst)
        with open(op.join(dst_dir, 'Contents', 'info.plist'), 'wb') as f:
            f.write(info.encode())
        # Make icon - DesktopRuntime ensures that there is an icon and title
        iconfile = op.join(dst_dir, 'Contents', 'Resources', 'app.icns')
        if op.exists(iconfile):
            os.unlink(iconfile)  # remove first, since its a hard link!
        self._icon.write(iconfile)
    
    ## To implenent in subclasses
    
    def _get_system_version(self):
        """ Provide (tuple, path) info on a system-available version.
        """
        raise NotImplementedError()
    
    def _install_runtime(self, sys_path, dest_path):
        """ Install a local copy of the latest runtime. Called when needed.
        """
        raise NotImplementedError()


# This does not seem to do much, most of the time, but when things break it
# usually does give some usefull output!
class StreamReader(threading.Thread):
    """ Reads stdout of process and log

    This needs to be done in a separate thread because reading from a
    PYPE blocks.
    """
    def __init__(self, process):
        threading.Thread.__init__(self)

        self._process = process
        self._exit = False
        self.setDaemon(True)
        atexit.register(self.stop)

    def stop(self, wait=None):  # pragma: no cover
        self._exit = True
        if wait is not None:
            self.join(wait)

    def run(self):  # pragma: no cover
        msgs = []
        while not self._exit:
            time.sleep(0.001)
            # Get and clean msg
            msg = self._process.stdout.readline()  # <-- Blocks here
            if not msg:
                break  # Process dead
            if not isinstance(msg, str):
                msg = msg.decode(errors='ignore')
            msg = msg.rstrip()
            # Process the message
            msgs.append(msg)
            msgs[:-32] = []
            logger.debug('from runtime: ' + msg)

        if self._exit:
            return  # might be interpreter shutdown, don't print

        # Poll to get return code. Polling also helps to really
        # clean the process up
        while self._process.poll() is None:
            time.sleep(0.05)

        # Notify
        code = self._process.poll()
        if getattr(self._process, 'we_closed_it', False):
            logger.info('runtime process terminated from Python')
        elif not code:
            logger.info('runtime process stopped')
        else:  # Unexpected, provide more info
            logger.error('runtime process (0x%x) stopped (%i), stdout:\n%s' %
                          (id(self._process), code, '\n'.join(msgs)))


def find_osx_exe(app_id):
    """ Find the xxx.app of an application via its app id,
    se.g. 'com.google.Chrome'.
    """
    try:
        cmd = ['mdfind', 'kMDItemCFBundleIdentifier==%s' % app_id]
        return subprocess.check_output(cmd).decode(errors='ignore').rstrip()
    except (OSError, subprocess.CalledProcessError):
        pass


## Icon stuff


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def iconize(icon):
    """ Given a filename Icon object or None, return Icon object.
    """
    
    # Get default icon?
    if icon is None:
        icon = Icon()
        icon.from_bytes('.ico', default_icon)
    
    if isinstance(icon, Icon):
        pass
    elif isinstance(icon, str):
        if icon.startswith('_data/shared'):
            # Icon as an asset in Flexx' asset store
            if 'flexx' in sys.modules:
                from flexx.app import assets  # noqa
                bb = assets.get_data(icon.split('/', 2)[-1])
                icon = Icon()
                icon.from_bytes('.ico', bb)
        else:
            # Filename, url, base64 string - handled by Icon class
            icon = Icon(icon)
    else:
        raise ValueError('Icon must be an Icon, str, or None, not %r' %
                         type(icon))
    return icon



""" Test some loose functions
"""

from webruntime.util.testing import run_tests_if_main, raises, skipif


from webruntime._manage import versionstring
from webruntime import _expand_runtime_name

def test_versionstring():
    
    # Format
    assert versionstring('10') .count('.') == 1
    assert versionstring('10.1') .count('.') == 2
    assert versionstring('10.1.2') .count('.') == 3
    
    # Ignore empty parts
    assert versionstring('10..1') == versionstring('10.1')
    assert versionstring('10....1..2') == versionstring('10.1.2')
    assert versionstring('10 . 1') == versionstring('10.1')
    
    # Allow recursion
    assert versionstring('10.1') == versionstring(versionstring('10.1'))
    
    # Simple comparisons
    assert versionstring('10') > versionstring('9')
    assert versionstring('10.1') > versionstring('9.1')
    assert versionstring('1.10') > versionstring('1.9')
    assert versionstring('10.1') > versionstring('9.9')
    assert versionstring('10.1.2.3.4.5.6') > versionstring('10.1.2.3.4.5.5')
    
    # Slightly trickier
    assert versionstring('10.1.1') > versionstring('10.1')
    assert versionstring('10.1.0') > versionstring('10.1')
    
    # Suffixes
    assert versionstring('10.1') > versionstring('10.1.a')
    assert versionstring('10.1') > versionstring('10.1a')
    assert versionstring('10.1rc1') > versionstring('10.1a')
    assert versionstring('10.1rc1') > versionstring('10.1.a')
    assert versionstring('10.1rc2') > versionstring('10.1rc1')
    
    # Latest is special
    assert versionstring('latest') > versionstring('999.999.999')



def test_expand_runtime_name():
    assert 'nw-app' in _expand_runtime_name('app')
    assert 'firefox-app' in _expand_runtime_name('app')
    

run_tests_if_main()



"""
Main file for invoke tasks. We auto-import all modules here, so that
one can simply add tasks by adding files.
"""

import os

from invoke import Collection, Task

# Get root directory of the package
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(THIS_DIR)

# Init collection; invoke picks this up as the main "list of tasks"
ns = Collection()

# Automatically collect tasks from submodules
for fname in os.listdir(THIS_DIR):
    # Does this look like a module that we want?
    if fname.startswith('_') or not fname.endswith('.py'):
        continue
    modname = fname[:-3]
    # Import it
    m = __import__(modname, level=1, fromlist=[], globals=globals())
    # Collect all tasks and collections
    collections, tasks = {}, {}
    for name in dir(m):
        ob = getattr(m, name)
        if isinstance(ob, Task):
            tasks[name] = ob
        elif isinstance(ob, Collection):
            collections[name] = ob
    # Add collections
    for name, ob in collections.items():
        ns.add_collection(ob, name)
    # Add tasks that are not already in a collection
    for name, ob in tasks.items():
        add_task = True
        for c in collections.values():
            if ob in c.tasks.values():
                add_task = False
        if add_task:
            ns.add_task(ob, name)



import os
import sys
import os.path as op
import shutil

from invoke import task
from ._config import DOC_DIR, DOC_BUILD_DIR


@task(help=dict(clean='clear the doc output; start fresh',
                build='build html docs',
                show='show the docs in the browser.'))
def docs(ctx, clean=False, build=False, show=False, **kwargs):
    """ make API documentation
    """
    # Prepare
    
    if not (clean or build or show):
        sys.exit('Task "docs" must be called with --clean, --build or --show')
    
    if clean:
        sphinx_clean(DOC_BUILD_DIR)
    
    if build:
        sphinx_build(DOC_DIR, DOC_BUILD_DIR)
    
    if show:
        sphinx_show(os.path.join(DOC_BUILD_DIR, 'html'))


def sphinx_clean(build_dir):
    if op.isdir(build_dir):
        shutil.rmtree(build_dir)
    os.mkdir(build_dir)
    os.mkdir(os.path.join(build_dir, 'html'))
    print('Cleared build directory.')


def sphinx_build(src_dir, build_dir):
    import sphinx
    cmd = [ '-b', 'html',
            '-d', op.join(build_dir, 'doctrees'),
            src_dir,  # Source
            op.join(build_dir, 'html'),  # Dest
            ]
    
    if sphinx.version_info > (1, 7):
        import sphinx.cmd.build
        ret = sphinx.cmd.build.build_main(cmd)
    else:
        ret = sphinx.build_main(['sphinx-build'] + cmd)
    if ret != 0:
        raise RuntimeError('Sphinx error: %s' % ret)
    print("Build finished. The HTML pages are in %s/html." % build_dir)


def sphinx_show(html_dir):
    index_html = op.join(html_dir, 'index.html')
    if not op.isfile(index_html):
        sys.exit('Cannot show pages, build the html first.')
    import webbrowser
    webbrowser.open_new_tab(index_html)



"""
Make this module itself executable as an alias for invoke.
"""

import sys
import subprocess

cmd = ['invoke']
if len(sys.argv) == 1:
    cmd.append('help')
else:
    cmd.extend(sys.argv[1:])

subprocess.check_call(cmd)



""" Web runtime based on a common browser

Opens browser via webbrowser module.
"""

import webbrowser

from . import logger
from ._common import BaseRuntime


BROWSER_MAP = {'chrome': ['google-chrome', 'chrome', 
                          'chromium-browser', 'chromium'],
               'chromium': ['chromium-browser', 'chromium'],
               'default': [],
               }


class BrowserRuntime(BaseRuntime):
    """ Runtime based on the Python webbrowser module. This runtime is used
    to open a url in the system default browser, and to attempt to handle a
    given browser name that is unknown to the webruntime module (maybe
    the webbrowser module knows it).
    """
    
    def __init__(self, type=None, **kwargs):
        self._type = type or ''
        super().__init__(**kwargs)
    
    def _get_install_instuctions(self):
        return ''  # we know nothing of the browsers
    
    def _get_name(self):
        return 'browser'
    
    def _get_exe(self):
        b, errors = self._get_openers(self._type)
        if not self._type:
            return 'stub_exe_default_browser'
        elif b:
            return 'stub_exe_%s_browser' % self._type
    
    def _get_version(self):
        return None
    
    def _launch_tab(self, url):
        
        b, errors = self._get_openers(self._type)
        if b:
            b.open(url)
        else:
            if errors:
                logger.warn('Given browser %r not valid/available;\n'
                            'Falling back to the default browser.' % type)
            # Run default
            webbrowser.open(url)
    
    def _launch_app(self, url):
        raise RuntimeError('Browser runtime cannot run as an app.')
    
    def _get_openers(self, type):
        # Get alternative types
        types = BROWSER_MAP.get(type, [type])
        types = [t for t in types if t]
        
        # Use default browser
        if not types:
            return webbrowser, []
        
        # Try to open all possibilities
        errors = []
        b = None
        for t in types:
            try:
                b = webbrowser.get(t)
                break
            except webbrowser.Error as err:
                errors.append(str(err))
        return b, errors



# -*- coding: utf-8 -*-
# Copyright (c) 2015-2017, Almar Klein
# This module is distributed under the terms of the new BSD License.

"""
Pure python module to handle the .ico format. Written for Python 2.7
and Python 3.2+. Depends on png.py.
"""

from __future__ import print_function, division, absolute_import

import sys
import struct

from .png import read_png, write_png

if sys.version_info[0] >= 3:
    basestring = str  # noqa
    from base64 import decodebytes
else:
    from base64 import decodestring as decodebytes


# Note: up to 256 is support by our .ico exporter
VALID_SIZES = 16, 32, 48, 64, 128, 256, 512, 1024

# Little endian int encoding (for bmp/icon writing)
w1 = lambda x: struct.pack('<B', x)
w2 = lambda x: struct.pack('<H', x)
w4 = lambda x: struct.pack('<I', x)


def intl(x):
    """ little endian int decoding (for bmp/ico reading)
    """
    if len(x) == 1:
        return struct.unpack('<B', x)[0]
    elif len(x) == 2:
        return struct.unpack('<H', x)[0]
    elif len(x) == 4:
        return struct.unpack('<I', x)[0]


class Icon(object):
    """
    Object for reading/creating icons. Considers only RGBA icons. Can
    deal with images stored raw (BMP) and compressed (PNG).
    """
    # Conventions:
    # im -> an image stores as a bytearray array, uint8, NxNx4
    # bb -> bytes, possibly a png/bmp/ico
    
    def __init__(self, *filenames):
        self._ims = {}
        for filename in filenames:
            self.read(filename)
    
    def __repr__(self):
        ss = self.image_sizes()
        return '<Icon with %i sizes: %r at 0x%x>' % (len(ss), ss, id(self))
    
    def image_sizes(self):
        """ Get a tuple of image sizes (integers) currently loaded.
        """
        return tuple(sorted(self._ims.keys()))
    
    def add(self, data):
        """ Add an image represented as bytes or bytearray. The size
        of the image is inferred from the number of bytes. The image
        is assumed to be square and in RGBA format.
        """
        if isinstance(data, (bytes, bytearray)):
            self._store_image(data)
        else:
            raise ValueError('Data to add should be bytes or bytearray')
    
    def read(self, filename):
        """ Read an image from the given filename and add to collection.
        Can be an ICO, PNG, or BMP file.  When a grayscale or RGB image
        is read, it is converted to RGBA. Some restrictions may apply
        to the formats that can be read.
        """
        if not isinstance(filename, basestring):
            raise TypeError('Icon.read() needs a file name')
        
        if filename.startswith(('http://', 'https://')):
            # Remote resource
            try:
                from urllib.request import urlopen  # Python 3.x
            except ImportError:
                from urllib2 import urlopen  # Python 2.x
            data = urlopen(filename, timeout=2.0).read()
        elif filename.startswith('data:image/') and 'base64' in filename[:32]:
            # Base64 encoded asset
            data = decodebytes(filename.split(',', 1)[-1].encode())
            filename = '.' + filename.split(';')[0].split('/')[-1]
        else:
            data = open(filename, 'rb').read()
        
        self.from_bytes(filename, data)
    
    def from_bytes(self, ext, data):
        """ Read an image from the raw bytes of the encoded image. The format
        is specified by the extension (or filename).
        """
        if ext.lower().endswith('.ico'):
            self._from_ico(data)
        elif ext.lower().endswith('.png'):
            self._from_png(data)
        elif ext.lower().endswith('.bmp'):
            self._from_bmp(data)
        else:
            raise ValueError('Can only load from png, bmp, or ico')
    
    def write(self, filename):
        """ Write the icon collection to an image with the given filename.
        Can be an ICO, ICNS, PNG, or BMP file. In case of PNG/BMP,
        multiple images may be generated, the image size is appended
        to the file name.
        """
        if not isinstance(filename, basestring):
            raise TypeError('Icon.write() needs a file name')
        
        if filename.lower().endswith('.ico'):
            data = self._to_ico()
            with open(filename, 'wb') as f:
                f.write(data)
        elif filename.lower().endswith('.icns'):
            data = self._to_icns()
            with open(filename, 'wb') as f:
                f.write(data)
        elif filename.lower().endswith('.png'):
            for size in sorted(self._ims):
                filename2 = '%s%i%s' % (filename[:-4], size, filename[-4:])
                data = self._to_png(self._ims[size])
                with open(filename2, 'wb') as f:
                    f.write(data)
        elif filename.lower().endswith('.bmp'):
            for size in sorted(self._ims):
                filename2 = '%s%i%s' % (filename[:-4], size, filename[-4:])
                data = self._to_bmp(self._ims[size], file_header=True)
                with open(filename2, 'wb') as f:
                    f.write(data)
        else:
            raise ValueError('Can only export to png, bmp, or ico')
    
    def to_bytes(self):
        """ Return the bytes that represent the .ico image.
        This function can be used by webservers to serve the ico image
        without needing a physical representation on disk.
        """
        return self._to_ico()
    
    def _image_size(self, im):
        npixels = len(im) // 4
        width = height = int(npixels ** 0.5)
        if width * height * 4 != len(im):
            raise ValueError('Icon must be NxMx4 pixels')
        if width not in VALID_SIZES:
            raise ValueError('Icon must have size in %s' % str(VALID_SIZES))
        return width
    
    def _store_image(self, im):
        self._ims[self._image_size(im)] = im
    
    def _from_ico(self, bb):
        # Windows icon format.
        # http://en.wikipedia.org/wiki/ICO_%28file_format%29
        
        assert intl(bb[0:2]) == 0
        assert intl(bb[2:4]) == 1  # must be ICO (not CUR)
        number_of_images = intl(bb[4:6])
        
        for imnr in range(number_of_images):
            imheader = bb[6+imnr*16:]
            # We don't care about dimensions and bpp, we read that in bmp/png
            width = intl(imheader[0:1]) or 256
            size = intl(imheader[8:12])
            offset = intl(imheader[12:16])
            # Get image
            imdata = bb[offset:offset+size]
            try:
                if imdata[1:4] == b'PNG':
                    self._from_png(imdata)
                else:
                    self._from_bmp(imdata)
            except RuntimeError as err:
                print('Skipping image size %i: %s' % (width, err))
    
    def _to_ico(self):
        
        bb = b''
        imdatas = []
        
        # Header
        bb += w2(0)
        bb += w2(1)  # 1:ICO, 2:CUR
        bb += w2(len(self._ims))
        
        # Put offset right after the last directory entry
        offset = len(bb) + 16 * len(self._ims)
        
        # Directory (header for each image)
        for size in sorted(self._ims):
            im = self._ims[size]
            if size > 256:
                continue
            elif size >= 64:
                imdata = self._to_png(im)
            else:
                imdata = self._to_bmp(im)
            imdatas.append(imdata)
            # Prepare dimensions
            w = h = 0 if size == 256 else size
            # Write directory entry
            bb += w1(w)
            bb += w1(h)
            bb += w1(0)  # number of colors in palette, assume no palette (0)
            bb += w1(0)  # reserved (must be 0)
            bb += w2(0)  # color planes
            bb += w2(32)  # bits per pixel
            bb += w4(len(imdata))  # size of image data
            bb += w4(offset)
            # Set offset pointer
            offset += len(imdata)
        
        if not imdatas:
            raise RuntimeError('Exported icon is empty '
                               '(none of the sizes was supported).')
        
        return b''.join([bb] + imdatas)
    
    def _to_icns(self):
        # OSX icon format. 
        # No formal spec. Any docs is reverse engineered by someone.
        # Which is one reason for not having a from_icns().
        # http://en.wikipedia.org/wiki/Apple_Icon_Image_format
        # http://www.macdisk.com/maciconen.php
        # http://www.ezix.org/project/wiki/MacOSXIcons
        
        imdatas = []
        raw_types = {16: (b'is32', b's8mk'),
                     32: (b'il32', b'l8mk'),
                     48: (b'ih32', b'h8mk'),
                     128: (b'it32', b't8mk'), }
        png_types = {16: b'icp4', 32: b'icp5', 64: b'icp6', 128: b'ic07',
                     256: b'ic08', 512: b'ic09', 1024: b'ic10'}
        
        for size in sorted(self._ims):
            im = self._ims[size]
            if size in raw_types:
                # Raw format - can be compressed with packbits
                type, apha_type = raw_types[size]
                # RGBA to XRGB
                data = bytearray(len(im))
                data[1::4] = im[0::4]
                data[2::4] = im[1::4]
                data[3::4] = im[2::4]
                # Store RGBA
                imdatas.append(type)
                imdatas.append(struct.pack('>I', len(data) + 8))
                imdatas.append(bytes(data))
                # RGBA to A
                data = bytearray(len(im)//4)
                data[:] = im[3::4]
                # Store alpha
                imdatas.append(apha_type)
                imdatas.append(struct.pack('>I', len(data) + 8))
                imdatas.append(bytes(data))
            elif False:  # size in png_types:
                # Store as png, does not seem to work
                data = self._to_png(im)
                imdatas.append(png_types[size])
                imdatas.append(struct.pack('>I', len(data) + 8))
                imdatas.append(bytes(data))
            else:
                # print('Skipping export size %i to .icns' % size)
                continue
        
        total_icon_size = sum([len(i) for i in imdatas]) + 8
        bb = b'icns' + struct.pack('>I', total_icon_size)
        return b''.join([bb] + imdatas)
    
    def _from_bmp(self, bb):
        # Bitmap image file format
        # http://en.wikipedia.org/wiki/BMP_file_format
        
        # Skip header it is there
        file_header = False
        if bb[0:2] in (b'BM', b'BA', b'CI', b'CP', b'IC', b'PT'):
            file_header = True
            bb = bb[14:]
        
        # Get and check header size (BMP identifies its many headers by length)    
        head_size = intl(bb[0:4])
        if head_size != 40:
            raise RuntimeError('Need BMP header size 40 (not %r)' % head_size)
        
        # Get info from this header
        width = intl(bb[4:8])
        height = intl(bb[8:12]) // (2 - file_header)  # half if not from file 
        #color_planes = intl(bb[12:14])
        bpp = intl(bb[14:16])
        compression = intl(bb[16:20])
        data_length = intl(bb[20:24])
        
        # Check
        if width != height:
            raise RuntimeError('Width and height must be equal in icon')
        if width not in VALID_SIZES:
            raise RuntimeError('Invalid size %r in icon' % width)
        if compression != 0:
            raise RuntimeError('Can only deal with uncompressed BMP')
        
        # Get image data
        im = bb[40:40+data_length]
        
        # Init, ensure we have alpha channel
        if bpp == 32:
            im = im[:width*width*4]  # Discart AND mask
            assert len(im) == width * width * 4
            im2 = bytearray(len(im))
            im2[3::4] = im[3::4]
        elif bpp == 24:
            im = im[width*width*3]  # Discart AND mask
            assert len(im) in width * width * 3
            im2 = bytearray(int(len(im)*1.333333333334))
            im2[3::4] = 255
        else:
            raise RuntimeError('Can only deal with RGB or RGBA BMP')
        
        # BGRA to RGBA
        im2[0::4] = im[2::4]
        im2[1::4] = im[1::4]
        im2[2::4] = im[0::4]
        im = im2
        
        # Flip vertically
        lines = [im[width*4*i:width*4*(i+1)] for i in range(height)]
        im = bytearray().join(reversed(lines))
        
        #return im2
        self._store_image(im)
        
    def _to_bmp(self, im, file_header=False):
        
        # Init
        width = self._image_size(im)
        height = reported_height = width
        if not file_header:
            reported_height *= 2  # This is soo weird, but it needs to be so
        
        # RGBA to BGRA
        im2 = bytearray(len(im))
        im2[0::4] = im[2::4]
        im2[1::4] = im[1::4]
        im2[2::4] = im[0::4]
        im2[3::4] = im[3::4]
        im = im2
        
        # Flip vertically
        lines = [im[width*4*i:width*4*(i+1)] for i in range(height)]
        im = bytearray().join(reversed(lines))
        
        # DIB header
        bb = b''
        bb += w4(40)  # header size
        bb += w4(width)
        bb += w4(reported_height)
        bb += w2(1)  # 1 color plane
        bb += w2(32)
        bb += w4(0)  # no compression
        bb += w4(len(im))
        bb += w4(2835) + w4(2835)  # 2835 pixels/meter, ~ 72 dpi
        bb += w4(0)  # number of colors in palette
        bb += w4(0)  # number of important colors (0->all)
        
        # File header (not when bm is in-memory)
        header = b''
        if file_header:
            header += b'BM'
            header += w4(14 + 40 + len(im))  # file size
            header += b'\x00\x00\x00\x00'
            header += w4(14 + 40)  # pixel data offset
        
        # Add pixels
        # No padding, because we assume power of 2 image sizes
        return header + bb + bytes(im)
    
    def _from_png(self, data):
        im, shape = read_png(data)
        
        if shape[0] != shape[1]:
            raise RuntimeError('Width and height must be equal in icon')
        if shape[0] not in VALID_SIZES:
            raise RuntimeError('Invalid size %r in png' % shape[0])
        
        # Make RGBA if necessary
        if shape[2] == 3:
            im2 = bytearray(int(len(im)*1.333333333334))
            im2[3::4] = b'\xff' * (shape[0] * shape[1])
            im2[0::4] = im[0::3]
            im2[1::4] = im[1::3]
            im2[2::4] = im[2::3]
        else:
            im2 = im  # already bytearray
        
        #return im2
        self._store_image(im2)
    
    def _to_png(self, im):
        size = self._image_size(im)
        return write_png(bytes(im), (size, size, 4))



import os
import sys

from invoke import task

from ._config import ROOT_DIR, NAME


@task
def lint(ctx):
    """ alias for "invoke test --style"
    """
    test_style()


@task(optional=['unit', 'style'],
      help=dict(unit='run unit tests (pytest) on given subdir (default ".")',
                style='run style tests (flake8) on given subdir (default ".")',
                cover='show test coverage'))
def test(ctx, unit='', style='', cover=False):
    """ run tests (unit, style)
    """

    if not (unit or style or cover):
        sys.exit('Test task needs --unit, --style or --cover')
    if unit:
        test_unit('.' if not isinstance(unit, str) else unit)
    if style:
        test_style('.' if not isinstance(style, str) else style)
    if cover:
        show_coverage_html()


def test_unit(rel_path='.'):
    # Ensure we have pytest
    try:
        import pytest  # noqa
    except ImportError:
        sys.exit('Cannot do unit tests, pytest not installed')
    # Get path to test
    py2 = sys.version_info[0] == 2
    rel_path = NAME + '/' + rel_path
    test_path = os.path.join(ROOT_DIR, rel_path)
    # Import from installed, or from ROOT_DIR
    if py2 or os.getenv('TEST_INSTALL', '').lower() in ('1', 'yes', 'true'):
        if ROOT_DIR in sys.path:
            sys.path.remove(ROOT_DIR)
        os.chdir(os.path.expanduser('~'))
        m = __import__(NAME)
        assert ROOT_DIR not in os.path.abspath(m.__path__[0])
    else:
        os.chdir(ROOT_DIR)
        m = __import__(NAME)
        assert ROOT_DIR in os.path.abspath(m.__path__[0])
    # Start tests
    _enable_faulthandler()
    try:
        res = pytest.main(['--cov', NAME, '--cov-config=.coveragerc',
                           '--cov-report=term', '--cov-report=html', test_path])
        sys.exit(res)
    finally:
        m = __import__(NAME)
        print('Unit tests were performed on', str(m))


def show_coverage_term():
    from coverage import coverage
    cov = coverage(auto_data=False, branch=True, data_suffix=None,
                   source=[NAME])  # should match testing/_coverage.py
    cov.load()
    cov.report()


def show_coverage_html():
    import webbrowser
    from coverage import coverage

    print('Generating HTML...')
    os.chdir(ROOT_DIR)
    cov = coverage(auto_data=False, branch=True, data_suffix=None,
                   source=[NAME])  # should match testing/_coverage.py
    cov.load()
    cov.html_report()
    print('Done, launching browser.')
    fname = os.path.join(os.getcwd(), 'htmlcov', 'index.html')
    if not os.path.isfile(fname):
        raise IOError('Generated file not found: %s' % fname)
    webbrowser.open_new_tab(fname)


def test_style(rel_path='.'):
    # Ensure we have flake8
    try:
        import flake8  # noqa
        from flake8.main.application import Application
    except ImportError as err:
        sys.exit('Cannot do style test: ' + str(err))
    # Prepare
    os.chdir(ROOT_DIR)
    # Do test
    print('Running flake8 tests ...')
    app = Application()
    app.run([NAME + '/' + rel_path])
    # Report
    nerrors = app.result_count
    if nerrors:
        print('Arg! Found %i style errors.' % nerrors)
    else:
        print('Hooray, no style errors found!')
    # Exit (will exit(1) if errors)
    app.exit()


def _enable_faulthandler():
    """ Enable faulthandler (if we can), so that we get tracebacks
    on segfaults.
    """
    try:
        import faulthandler
        faulthandler.enable()
        print('Faulthandler enabled')
    except Exception:
        print('Could not enable faulthandler')



import os
import shutil
import fnmatch

from invoke import task

from ._config import ROOT_DIR, NAME


@task
def clean(ctx):
    """ clear all .pyc modules and __pycache__ dirs
    """
    count1, count2 = 0, 0
    
    for root, dirnames, filenames in os.walk(ROOT_DIR):
        for dirname in dirnames:
            if dirname == '__pycache__':
                shutil.rmtree(os.path.join(root, dirname))
                count1 += 1
    print('removed %i __pycache__ dirs' % count1)
    
    for root, dirnames, filenames in os.walk(ROOT_DIR):
        for filename in fnmatch.filter(filenames, '*.pyc'):
            os.remove(os.path.join(root, filename))
            count2 += 1
    print('removed %i .pyc files' % count2)
    
    for dir in ['dist', 'build', NAME+'.egg-info', 'htmlcov']:
        dirname = os.path.join(ROOT_DIR, dir)
        if os.path.isdir(dirname):
            shutil.rmtree(dirname)
            print('Removed directory %r' % dir)



""" Web runtime based on Selenium.

Selenium is a Python library to automate browsers. 

"""

from ._common import BaseRuntime


class SeleniumRuntime(BaseRuntime):
    """ Runtime based on Selenium (http://www.seleniumhq.org/), a tool
    to automate browsers, e.g. for testing. Requires the Python package
    "selenium" to be installed. It's possible that this runtime is currently
    broken.
    """
    
    def __init__(self, type=None, **kwargs):
        self._type = type or ''
        super().__init__(**kwargs)
    
    def _get_install_instuctions(self):
        return ('To enable the Selenium runtime, install selenium '
                'in your Python environment.')
    
    def _get_name(self):
        return 'selenium'
    
    def _get_exe(self):
        try:
            import selenium
            return selenium.__file__
        except ImportError:
            return None
    
    def _get_version(self):
        if self._get_exe():
            import selenium
            return selenium.__version__
    
    def _launch_tab(self, url):
        
        # Get url and browser type
        type = self._type
        self._driver = None
        
        # Import here; selenium is an optional dependency
        from selenium import webdriver
        
        if type.lower() == 'firefox':
            self._driver = webdriver.Firefox()
        elif type.lower() == 'chrome':
            self._driver = webdriver.Chrome()
        elif type.lower() == 'ie':
            self._driver = webdriver.Ie()
        elif type:
            classname = None
            type2 = type[0].upper() + type[1:]
            if hasattr(webdriver, type):
                classname = type
            elif hasattr(webdriver, type2):
                classname = type2
            
            if classname:
                self._driver = getattr(webdriver, classname)()
            else:
                raise ValueError('Unknown Selenium browser type %r' % type)
            
        else:
            raise ValueError('To use selenium runtime specify a browser type".')
        
        # Open page
        self._driver.get(url)
    
    def _launch_app(self, url):
        raise RuntimeError('Selenium runtime cannot run as an app.')
    
    def close(self):
        """ Close the Selenium driver.
        """
        if self._driver:
            self._driver.close()
            self._driver = None
    
    def get_driver(self):
        """ Get the Selenium webdriver object. Use this to control the browser.
        """
        return self._driver




import os
import time
import tempfile
import subprocess

from webruntime.util import icon

from webruntime.util.testing import run_tests_if_main, raises, skipif

from webruntime import BaseRuntime
from webruntime import launch
import webruntime


URL = None

HTML = """
<html>
<head>
<meta charset="utf-8">
<style>
    body {background: #00aa00;}
</style>
</head
<body></body>
</html>
"""


def setup_module():
    global URL, FILE
    fname = os.path.join(tempfile.gettempdir(), 'webruntime_testpage.html')
    with open(fname, 'wb') as f:
        f.write(HTML.encode())
    URL = 'file://' + fname


def has_qt():
    try:
        from PyQt4 import QtWebKit
    except ImportError:
        try:
            from PySide import QtWebKit
        except ImportError:
            return False
    return True


def has_nw():
    return webruntime.NWRuntime().is_available()


def has_chrome():
    return webruntime.ChromeRuntime().is_available()


## Misc

def test_iconize():

    # Default icon
    icn = webruntime._common.iconize(None)
    assert isinstance(icn, icon.Icon)

    fname = os.path.join(tempfile.gettempdir(), 'webruntime_testicon.ico')
    icn.write(fname)

    # Load from file
    icn = webruntime._common.iconize(fname)
    assert isinstance(icn, icon.Icon)

    # Load from icon (noop)
    assert webruntime._common.iconize(icn) is icn

    # Error
    raises(ValueError, webruntime._common.iconize, [])


## Runtimes


@skipif(not has_qt(), reason='need qt')
def test_qtwebkit():
    p = launch(URL, 'pyqt-app')
    assert p._proc
    p.close()


def test_xul():
    p = launch(URL, 'firefox-app')
    assert p._proc

    p.close()
    p.close()  # should do no harm


@skipif(not has_nw(), reason='need nw')
def test_nwjs():
    p = launch(URL, 'nw-app')
    assert p._proc
    p.close()


@skipif(not has_chrome(), reason='need chrome/chromium')
def test_chomeapp():
    p = launch(URL, 'chrome-app')
    assert p._proc
    p.close()


def test_browser():
    p = launch(URL, 'default-browser')
    assert p._proc is None


def test_browser_ff():
    p = launch(URL, 'firefox-browser')
    assert p._proc is None


#@skipif(os.getenv('TRAVIS') == 'true', reason='skip selenium on Travis')
@skipif(True, reason='meh selenium')
def test_selenium():
    p = launch(URL, 'selenium-firefox')
    assert p._proc is None
    assert p.driver
    time.sleep(0.5)
    p.close()
    raises(ValueError, launch, URL, 'selenium')


def test_unknown():
    # Suppress dialog temporarily
    import dialite
    with dialite.NoDialogs():
        raises(ValueError, launch, URL, 'foo')


def test_default():
    p = launch(URL)
    assert p.__class__.__name__ == 'FirefoxRuntime'
    p.close()


run_tests_if_main()



# -*- coding: utf-8 -*-
# Copyright (c) 2016, Almar Klein
# Distributed under the (new) BSD License.

"""
Functionality used for testing, based on pytest. This module is designed
to just work, without modification, in most projects.

Write your tests like this:

    from yourproject.xxx.testing import run_tests_if_main, raises, skipif
    ...
    run_tests_if_main()

Then you can run the test file as a script, which will run all tests
and report coverage. Magic!
"""


from __future__ import absolute_import, print_function, division

import os
import sys
import inspect

import pytest


PACKAGE_NAME = __name__.split('.')[0]

# Get project root dir
THIS_DIR = os.path.abspath(os.path.dirname(__file__))
ROOT_DIR = THIS_DIR
for i in range(9):
    ROOT_DIR = os.path.dirname(ROOT_DIR)
    if os.path.basename(ROOT_DIR) == PACKAGE_NAME:
        ROOT_DIR = os.path.dirname(ROOT_DIR)
        break
else:
    print('testing.py could not find project root dir, '
          'using testing.py directory instead.')
    ROOT_DIR = THIS_DIR


# Inject some function names so they can be obtained with one import
raises = pytest.raises
skipif = pytest.mark.skipif
skip = pytest.skip


def run_tests_if_main(show_coverage=False):
    """ Run tests in a given file if it is run as a script
    
    Coverage is reported for running this single test. Set show_coverage to
    launch the report in the web browser.
    """
    local_vars = inspect.currentframe().f_back.f_locals
    if not local_vars.get('__name__', '') == '__main__':
        return
    # we are in a "__main__"
    os.chdir(ROOT_DIR)
    fname = str(local_vars['__file__'])
    _clear_our_modules()
    _enable_faulthandler()
    pytest.main(['-v', '-x', '--color=yes', '--cov', PACKAGE_NAME,
                 '--cov-config', '.coveragerc', '--cov-report', 'html', fname])
    if show_coverage:
        import webbrowser
        fname = os.path.join(ROOT_DIR, 'htmlcov', 'index.html')
        webbrowser.open_new_tab(fname)


def _enable_faulthandler():
    """ Enable faulthandler (if we can), so that we get tracebacks
    on segfaults.
    """
    try:
        import faulthandler
        faulthandler.enable()
        print('Faulthandler enabled')
    except Exception:
        print('Could not enable faulthandler')


def _clear_our_modules():
    """ Remove ourselves from sys.modules to force an import.
    """
    for key in list(sys.modules.keys()):
        if key.startswith(PACKAGE_NAME) and 'testing' not in key:
            del sys.modules[key]



import os

from ._config import ROOT_DIR

from invoke import task


@task
def copyright(ctx):
    """ list usage of copyright notices
    
    The use of copyright notices should be limited to files that are likely
    to be used in other projects, or to make appropriate attributions for code
    taken from other projects. Other than that, git geeps track of what person
    wrote what.
    """
    
    # Processing the whole root directory
    for dirpath, dirnames, filenames in os.walk(ROOT_DIR):
        # Check if we should skip this directory
        reldirpath = os.path.relpath(dirpath, ROOT_DIR)
        if reldirpath[0] in '._' or reldirpath.endswith('__pycache__'):
            continue
        if os.path.split(reldirpath)[0] in ('build', 'dist'):
            continue
        # Process files
        for fname in filenames:
            if not fname.endswith('.py'):
                continue
            # Open and check
            filename = os.path.join(dirpath, fname)
            text = open(filename, 'rt', encoding='utf-8').read()
            if 'copyright' in text[:200].lower():
                print(
                    'Copyright in %s%s%s' % (reldirpath, os.path.sep, fname))
                for i, line in enumerate(text[:200].splitlines()):
                    if 'copyright' in line.lower():
                        print('  line %i: %s' % (i+1, line))



""" Make screenshots of windows on Windows and Linux.
We need this to do visual tests.
"""

import sys

if sys.platform.startswith('win'):
    
    import ctypes
    from ctypes import windll
    from ctypes.wintypes import (BOOL, DOUBLE, DWORD, HBITMAP, HDC, HGDIOBJ,  # noqa
                                 HWND, INT, LPARAM, LONG, UINT, WORD)  # noqa
    
    SRCCOPY = 13369376
    DIB_RGB_COLORS = BI_RGB = 0
    
    class RECT(ctypes.Structure):
        _fields_ = [('left', ctypes.c_long),
                    ('top', ctypes.c_long),
                    ('right', ctypes.c_long),
                    ('bottom', ctypes.c_long)]
    
    class BITMAPINFOHEADER(ctypes.Structure):
        _fields_ = [('biSize', DWORD), ('biWidth', LONG), ('biHeight', LONG),
                    ('biPlanes', WORD), ('biBitCount', WORD),
                    ('biCompression', DWORD), ('biSizeImage', DWORD),
                    ('biXPelsPerMeter', LONG), ('biYPelsPerMeter', LONG),
                    ('biClrUsed', DWORD), ('biClrImportant', DWORD)]
    
    class BITMAPINFO(ctypes.Structure):
        _fields_ = [('bmiHeader', BITMAPINFOHEADER), ('bmiColors', DWORD * 3)]
    
    # Function shorthands
    GetClientRect = windll.user32.GetClientRect
    GetWindowRect = windll.user32.GetWindowRect
    PrintWindow = windll.user32.PrintWindow
    GetWindowThreadProcessId = windll.user32.GetWindowThreadProcessId
    IsWindowVisible = windll.user32.IsWindowVisible
    EnumWindows = windll.user32.EnumWindows
    EnumWindowsProc = ctypes.WINFUNCTYPE(ctypes.c_bool,
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int))
    
    GetWindowDC = windll.user32.GetWindowDC
    CreateCompatibleDC = windll.gdi32.CreateCompatibleDC
    CreateCompatibleBitmap = windll.gdi32.CreateCompatibleBitmap
    SelectObject = windll.gdi32.SelectObject
    BitBlt = windll.gdi32.BitBlt
    DeleteObject = windll.gdi32.DeleteObject
    GetDIBits = windll.gdi32.GetDIBits
    
    # Arg types
    windll.user32.GetWindowDC.argtypes = [HWND]
    windll.gdi32.CreateCompatibleDC.argtypes = [HDC]
    windll.gdi32.CreateCompatibleBitmap.argtypes = [HDC, INT, INT]
    windll.gdi32.SelectObject.argtypes = [HDC, HGDIOBJ]
    windll.gdi32.BitBlt.argtypes = [HDC, INT, INT, INT, INT, HDC, INT, INT, DWORD]
    windll.gdi32.DeleteObject.argtypes = [HGDIOBJ]
    windll.gdi32.GetDIBits.argtypes = [HDC, HBITMAP, UINT, UINT, ctypes.c_void_p,
                                        ctypes.POINTER(BITMAPINFO), UINT]
    # Return types
    windll.user32.GetWindowDC.restypes = HDC
    windll.gdi32.CreateCompatibleDC.restypes = HDC
    windll.gdi32.CreateCompatibleBitmap.restypes = HBITMAP
    windll.gdi32.SelectObject.restypes = HGDIOBJ
    windll.gdi32.BitBlt.restypes = BOOL
    windll.gdi32.GetDIBits.restypes = INT
    windll.gdi32.DeleteObject.restypes = BOOL
    
    
    def win_for_pid(pid):
        """ Get the windows-handle for the first visible window of the
        process with the given id.
        """
        handles = []
        
        def called_for_each_win(hwnd, lParam):
            if not IsWindowVisible(hwnd):
                return True
            # get the proccessid from the windowhandle
            p_id = ctypes.c_int()
            #t_id = GetWindowThreadProcessId(hwnd, ctypes.byref(p_id))
            if p_id.value == pid:
                handles.append(hwnd)
                return False
            return True
        
        EnumWindows(EnumWindowsProc(called_for_each_win), 0)
        if handles:
            return handles[0]
        else:
            return None
    
    
    def screenshot(pid, client=True):
        """ Grab a screenshot of the first visible window of the process
        with the given id. If client is True, no Window decoration is shown.
        
        This code is derived from https://github.com/BoboTiG/python-mss
        """
        # Get handle
        hwnd = win_for_pid(pid)
        # Get window dimensions
        rect = RECT()
        if client:
            GetClientRect(hwnd, ctypes.byref(rect))
        else:
            GetWindowRect(hwnd, ctypes.byref(rect))
        left, right, top, bottom = rect.left, rect.right, rect.top, rect.bottom
        w, h = right - left, bottom - top
        
        hwndDC = saveDC = bmp = None
        try:
            # Get device contexts
            hwndDC = GetWindowDC(hwnd)
            saveDC = CreateCompatibleDC(hwndDC)
            # Get bitmap
            bmp = CreateCompatibleBitmap(hwndDC, w, h)
            SelectObject(saveDC, bmp)
            if client:
                PrintWindow(hwnd, saveDC, 1)  # todo: result is never used??
            else:
                PrintWindow(hwnd, saveDC, 0)
            # Init bitmap info
            # We grab the image in RGBX mode, so that each word is 32bit and
            # we have no striding, then we transform to RGB
            buffer_len = h * w * 4
            bmi = BITMAPINFO()
            bmi.bmiHeader.biSize = ctypes.sizeof(BITMAPINFOHEADER)
            bmi.bmiHeader.biWidth = w
            bmi.bmiHeader.biHeight = -h  # Why minus? See [1]
            bmi.bmiHeader.biPlanes = 1  # Always 1
            bmi.bmiHeader.biBitCount = 32
            bmi.bmiHeader.biCompression = BI_RGB
            # Blit
            image = ctypes.create_string_buffer(buffer_len)
            bits = windll.gdi32.GetDIBits(saveDC, bmp, 0, h, image, bmi, DIB_RGB_COLORS)
            assert bits == h
            # Replace pixels values: BGRX to RGB
            image2 = ctypes.create_string_buffer(h*w*3)
            image2[0::3] = image[2::4]
            image2[1::3] = image[1::4]
            image2[2::3] = image[0::4]
            
            return bytes(image2), (w, h, 3)
        
        finally:
            # Clean up
            if hwndDC:
                DeleteObject(hwndDC)
            if saveDC:
                DeleteObject(saveDC)
            if bmp:
                DeleteObject(bmp)


if __name__ == '__main__':
    im, shape = screenshot(5144, True)
    
    from webruntime.util import icon
    png = icon.write_png(im, shape)
    open('C:\\Users\\Almar\\test.png', 'wb').write(png)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, Almar Klein
# This module is distributed under the terms of the new BSD License.

"""
Pure python module to handle for reading and writing png files. Written
for Python 2.7 and Python 3.2+. Can only read PNG's that are not
interlaced, have a bit depth of 8, and are either RGB or RGBA.
"""

from __future__ import print_function, division, absolute_import

import io
import struct
import zlib


def write_png(im, shape=None, file=None):
    """
    Write a png image. The written image is in RGB or RGBA format, with
    8 bit precision, and without interlacing.
    
    Parameters:
        im (bytes, bytearray, numpy-array): the image data to write.
        shape (tuple): the shape of the image. If ``im`` is a numpy array,
            the shape can be omitted. The shape can be ``(H, W)`` for
            grayscale, ``(H, W, 3)`` for RGB and ``(H, W, 4)`` for RGBA.
            Note that grayscale images are converted to RGB.
        file (file-like object, None): where to write the resulting
            image. If omitted or None, the result is returned as bytes.
    """
    
    # Check types
    if hasattr(im, 'shape') and hasattr(im, 'dtype'):
        if shape and tuple(shape) != im.shape:
            raise ValueError('write_png got mismatch in im.shape and shape')
        if im.dtype != 'uint8':
            raise TypeError('Image data to write to PNG must be uint8')
        shape = im.shape
        im = im.tobytes()
    elif isinstance(im, (bytes, bytearray)):
        if not isinstance(shape, (tuple, list)):
            raise ValueError('write_png needs a shape unless ndarray is given')
        shape = tuple(shape)
    else:
        raise ValueError('Invalid type for im, '
                         'need ndarray, bytearray or bytes, got %r' % type(im))
    
    # Allow grayscale: convert to RGB
    if len(shape) == 2 or (len(shape) == 3 and shape[2] == 1):
        im3 = bytearray(shape[0] * shape[1] * 3)
        im3[0::3] = im
        im3[1::3] = im
        im3[2::3] = im
        im = im3
        shape = shape[0], shape[1], 3
    
    # Check shape
    if len(shape) != 3:
        raise ValueError('shape must be 3 elements)')
    if shape[2] not in (3, 4):
        raise ValueError('shape[2] must be in (3, 4)')
    if (shape[0] * shape[1] * shape[2]) != len(im):
        raise ValueError('Shape does not match number of elements in image')
    
    # Get file object
    f = io.BytesIO() if file is None else file
    
    def add_chunk(data, name):
        name = name.encode('ASCII')
        crc = zlib.crc32(data, zlib.crc32(name))
        f.write(struct.pack('>I', len(data)))
        f.write(name)
        f.write(data)
        #f.write(crc.to_bytes(4, 'big'))  # python 3.x +
        f.write(struct.pack('>I', crc & 0xffffffff))
    
    f.write(b'\x89PNG\x0d\x0a\x1a\x0a')  # header
    
    # First chunk
    w, h = shape[1], shape[0]
    depth = 8
    ctyp = 0b0110 if shape[2] == 4 else 0b0010
    ihdr = struct.pack('>IIBBBBB', w, h, depth, ctyp, 0, 0, 0)
    add_chunk(ihdr, 'IHDR')
    
    # Chunk with pixels. Just one chunk, no fancy filters.
    line_len = w * shape[2]
    lines = [im[i*line_len:(i+1)*line_len] for i in range(h)]
    lines = [b'\x00' + bytes(line) for line in lines]  # prepend filter byt
    pixels_compressed = zlib.compress(b''.join(lines), 9)
    add_chunk(pixels_compressed, 'IDAT')
    
    # Closing chunk
    add_chunk(b'', 'IEND')
    
    if file is None:
        return f.getvalue()


def read_png(f, return_ndarray=False):
    """
    Read a png image. This is a simple implementation; can only read
    PNG's that are not interlaced, have a bit depth of 8, and are either
    RGB or RGBA.
    
    Parameters:
        f (file-object, bytes): the source to read the png data from.
        return_ndarray (bool): whether to return the result as a numpy array.
            Default False. If False, returns ``(pixel_array, shape)``,
            with ``pixel_array`` a bytearray object and shape being
            ``(H, W, 3)`` or ``(H, W, 4)``, for RGB and RGBA, respectively.
    """
    # http://en.wikipedia.org/wiki/Portable_Network_Graphics
    # http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html
    
    asint_map = {1: '>B', 2: '>H', 4: '>I'}
    asint = lambda x: struct.unpack(asint_map[len(x)], x)[0]
    
    # Get bytes
    if isinstance(f, (bytes, bytearray)):
        bb = f
    elif hasattr(f, 'read'):
        bb = f.read()
    else:
        raise TypeError('read_png() needs file object or bytes, not %r' % f)
    
    # Read header
    if not (bb[0:1] == b'\x89' and bb[1:4] == b'PNG'):
        raise RuntimeError('Image data does not appear to have a PNG '
                           'header: %r' % bb[:10])
    chunk_pointer = 8
    
    # Read first chunk
    chunk1 = bb[chunk_pointer:]
    chunk_length = asint(chunk1[0:4])
    chunk_pointer += 12 + chunk_length  # size, type, crc, data
    if not (chunk1[4:8] == b'IHDR' and chunk_length == 13):  # noqa
        raise RuntimeError('Unable to read PNG data, maybe its corrupt?')
    
    # Extract info
    width = asint(chunk1[8:12])
    height = asint(chunk1[12:16])
    bit_depth = asint(chunk1[16:17])
    color_type = asint(chunk1[17:18])
    compression_method = asint(chunk1[18:19])
    filter_method = asint(chunk1[19:20])
    interlace_method = asint(chunk1[20:21])
    bytes_per_pixel = 3 + (color_type == 6)
    
    # Check if we can do this ....
    if bit_depth != 8:
        raise RuntimeError('Can only deal with bit-depth of 8.')
    if color_type not in [2, 6]:  # RGB, RGBA
        raise RuntimeError('Can only deal with RGB or RGBA.')
    if interlace_method != 0:
        raise RuntimeError('Can only deal with non-interlaced.')
    if filter_method != 0:
        raise RuntimeError('Can only deal with unfiltered data.')
    if compression_method != 0:
        # this should be the case for any PNG
        raise RuntimeError('Expected PNG compression param to be 0.')
    
    # If this is the case ... extract pixel info
    lines, prev = [], None
    while True:
        chunk = bb[chunk_pointer:]
        if not chunk:
            break
        chunk_length = asint(chunk[0:4])
        chunk_pointer += 12 + chunk_length
        if chunk[4:8] == b'IEND':
            break
        elif chunk[4:8] == b'IDAT':  # Pixel data
            # Decompress and unfilter
            pixels_compressed = chunk[8:8+chunk_length]
            pixels_raw = zlib.decompress(pixels_compressed)
            s = width * bytes_per_pixel + 1  # stride
            #print(pixels_raw[0::s])  # show filters in use
            for i in range(height):
                prev = _png_scanline(pixels_raw[i*s:i*s+s], prev=prev)
                lines.append(prev)
            
    # Combine scanlines from all chunks
    im = bytearray(sum([len(line) for line in lines]))
    i = 0
    line_len = width * bytes_per_pixel
    for line in lines:
        if not len(line) == line_len:  # noqa
            raise RuntimeError('Line length mismatch while reading png.')
        im[i:i+line_len] = line
        i += line_len
    
    shape = height, width, bytes_per_pixel
    
    # Done
    if return_ndarray:
        import numpy as np
        return np.frombuffer(im, 'uint8').reshape(shape)
    else:
        return im, shape


def _png_scanline(line_bytes, fu=4, prev=None):
    """ Scanline unfiltering, taken from png.py
    """
    filter = ord(line_bytes[0:1])
    line1 = bytearray(line_bytes[1:])  # copy so that indexing yields ints
    line2 = bytearray(line_bytes[1:])  # output line
    
    if filter == 0:
        pass  # No filter
    elif filter == 1:
        # sub
        ai = 0
        for i in range(fu, len(line2)):
            x = line1[i]
            a = line2[ai]
            line2[i] = (x + a) & 0xff
            ai += 1
    elif filter == 2:
        # up
        for i in range(len(line2)):
            x = line1[i]
            b = prev[i]
            line2[i] = (x + b) & 0xff
    elif filter == 3:
        # average
        ai = -fu
        for i in range(len(line2)):
            x = line1[i]
            if ai < 0:
                a = 0
            else:
                a = line2[ai]
            b = prev[i]
            line2[i] = (x + ((a + b) >> 1)) & 0xff
            ai += 1
    elif filter == 4:
        # paeth
        ai = -fu  # Also used for ci.
        for i in range(len(line2)):
            x = line1[i]
            if ai < 0:
                a = c = 0
            else:
                a = line2[ai]
                c = prev[ai]
            b = prev[i]
            p = a + b - c
            pa = abs(p - a)
            pb = abs(p - b)
            pc = abs(p - c)
            if pa <= pb and pa <= pc:
                pr = a
            elif pb <= pc:
                pr = b
            else:
                pr = c
            line2[i] = (x + pr) & 0xff
            ai += 1
    else:
        raise RuntimeError('Invalid filter %r' % filter)
    return line2



"""
The webruntime module can be used to launch applications based on
HTML/JS/CSS. This can be a browser or a runtime that looks like a
desktop app, such as XUL (based on Firefox) or NW.js.

Example:

.. code-block:: python
    
    >>> from webruntime import launch
    >>> rt = launch('http://xkcd.com', 'app')
    ...
    >>> rt.close()

"""

"""
Memory considerations
---------------------

* xul uses one process, taking about 45 MB app
* pyqt uses one process, taking about 48 MB per app
* chrome uses 4 process plus 3 per app, taking 100 MB plus 74 per app

"""

__version__ = '0.5.8'


import sys
import logging
import traceback
from collections import OrderedDict
logger = logging.getLogger(__name__)
del logging

from ._config import config  # noqa
from ._manage import RUNTIME_DIR, TEMP_APP_DIR  # noqa
from ._common import BaseRuntime, DesktopRuntime  # noqa
from ._firefox import FirefoxRuntime
from ._nw import NWRuntime
from ._browser import BrowserRuntime
from ._qt import PyQtRuntime
from ._chrome import ChromeRuntime, GoogleChromeRuntime, ChromiumRuntime
from ._ms import IERuntime, EdgeRuntime
from ._selenium import SeleniumRuntime

import dialite


# Definition of all runtime names and their order
_runtimes = OrderedDict()
_runtimes['firefox'] = FirefoxRuntime
_runtimes['nw'] = NWRuntime
_runtimes['pyqt'] = PyQtRuntime
_runtimes['chrome'] = ChromeRuntime  # though handled via aliases
_runtimes['googlechrome'] = GoogleChromeRuntime
_runtimes['chromium'] = ChromiumRuntime
_runtimes['edge'] = EdgeRuntime
_runtimes['ie'] = IERuntime
_runtimes['browser'] = BrowserRuntime


_aliases = {'app': 'firefox-app or nw-app',
            'browser': ('firefox-browser or chrome-browser or edge-browser '
                        'or default-browser'),
            'chrome-browser': 'googlechrome-browser or chromium-browser',
            'chrome-app': 'googlechrome-app or chromium-app',
            }

if sys.platform.startswith('win'):
    _aliases['app'] = 'firefox-app or chrome-app or nw-app'

# We require to specify -app or -browser suffixes, though old names still work
_aliases_compat = {
            'xul': 'firefox-app',
            'nw': 'nw-app',
            'nwjs': 'nw-app',
            'chromeapp': 'chrome-app',
            'pyqt': 'pyqt-app',
            'firefox': 'firefox-browser',
            'chrome': 'chrome-browser',
}


def launch(url, runtime=None, **kwargs):
    """ Launch a web runtime in a new process.
    
    Parameters:
        url (str): The url to open, e.g. ``'http://python.org'``. To open a
            local file use ``'file://...'``.
        runtime (str) : The runtime(s) to use. E.g. 'app' will open in a
            desktop-app-like runtime, 'browser' in a browser runtime. One can
            target specific runtimes, e.g. 'nw-app' or 'edge-browser', or
            a selection, e.g. 'chrome-browser or firefox-browser'. Default
            is ``'app or browser'``.
            See below for more information on available runtimes.
        kwargs: addition arguments specific to the runtime. See the
            docs of the runtime classes.
    
    Returns:
        BaseRuntime: An object that represents the runtime. For
        Desktop runtimes it can be used to close the runtime.
    
    Browser runtimes:
    
    * browser: open in a browser. Firefox, Chrome and Edge are prefered over
      the default browser.
    * default-browser: open in the system default browser.
    * firefox-browser: open in Firefox browser.
    * chrome-browser: open in Chrome/Chromium browser.
    * googlechrome-browser or chromium-browser: like chrome-browser, but specific.
    * edge-browser: open in Microsoft Edge browser.
    * ie-browser: open in Microsoft Internet Explorer browser.
    * xx-browser: use webbrowser module to open in browser "xx".
    
    App runtimes:
    
    * app: open as desktop app, using firefox-app or nw-app
      (and chrome-app on Windows).
    * firerox-app: open as desktop app, using Firefox' app framework (Xul).
    * nw-app: open as desktop app using NW.js.
    * chrome-app: open as desktop-like app via Chrome/Chromium (only works well
      on Windows).
    * pyqt-app: open as desktop-like app using PyQt/PySide.
    
    The most developed app runtimes are Firefox and NW. The former requires
    the user to have Firefox installed. The latter can be installed by the user
    simply by downloading the archive. Firefox is lighter (memory-wise), while
    NW is based on Chromium, making it heavier. The other
    app runtimes are useful for testing or development, but should generally be
    avoided when distributing apps.
    
    """
    
    # Resolve backward compat names, and select default runtime if not given
    if runtime in _aliases_compat:
        logger.warn('Runtime name %s is deprecated, use %s instead.' %
                    (runtime, _aliases_compat[runtime]))
        runtime = _aliases_compat[runtime]
    if not runtime:
        runtime = 'app or browser'
    given_runtime = runtime
    
    # Normalize runtime, apply aliases
    runtimes = _expand_runtime_name(runtime)
    tried_runtimes = []
    errors = []
    
    # Attempt to launch runtimes, one by one
    for runtime in runtimes: 
        rt, launched, err = _launch(url, runtime, **kwargs)
        if rt and launched:
            return rt  # Hooray!
        if rt:
            tried_runtimes.append(rt)
        if err:
            errors.append(str(err).strip())
    
    # We end up here only when no suitable runtime was found.
    # Note that default-browser will always work, so by default we wont
    # end up here. We can well get here when runtime is 'app' though.
    
    # Show dialog to the user with information on how to install a runtime.
    # It is important that this is a dialog and not printed to stdout for
    # cases where an app is frozen (e.g. with cx_Freeze), because there is
    # no stdout in that case. Dialite will fallback to stdout if there is no
    # way to create a dialog, and if there is a tty, and attempt to show a
    # webpage with an error message otherwise.
    messages = []
    if not tried_runtimes:
        messages.append('Given runtime name "%s" does '
                        'not resolve to any known runtimes.' % given_runtime)
    elif len(tried_runtimes) == 1:
        # This app needs exactly this runtime
        rt = tried_runtimes[0]
        name = given_runtime if given_runtime.endswith('-browser') else rt.get_name()
        msg = 'Could not run app, because runtime %s ' % name
        msg += 'could not be used.' if errors else 'is not available.'
        messages.append(msg)
        if rt._get_install_instuctions():
            messages.append(rt._get_install_instuctions())
    else:
        # User has options
        seen = set()
        messages.append('Could not find a suitable runtime to run app. '
                        'Available options:')
        for c, rt in zip('ABCDEFGHIJK', tried_runtimes):
            if rt.get_name() in seen or not rt._get_install_instuctions():
                continue
            seen.add(rt.get_name())
            messages.append(c + ': ' + rt._get_install_instuctions())
    if errors:
        messages.append('Errors:')
        messages.extend(errors)
    messages = '\n\n'.join(messages)
    
    dialite.fail('Webruntime - No suitable runtime available', messages)
    
    raise ValueError('Could not detect a suitable backend among %r.' % runtimes)



def _launch(url, runtime, **kwargs):
    """ Attempt to launch runtime by its name.
    Return (runtime_object, is_launched, error_object)
    """
    
    rt = None
    launched = False
    
    try:
    
        if runtime.endswith('-app'):
            # Desktop-like app runtime
            runtime = runtime.split('-')[0]
            Runtime = _runtimes.get(runtime, None)
            if Runtime is None:
                logger.warn('Unknown app runtime %r.' % runtime)
            else:
                rt = Runtime(**kwargs)
                if rt.is_available():
                    rt.launch_app(url)
                    launched = True
        
        elif runtime.startswith('selenium-'):
            # Selenium runtime - always try or fail
            if '-' in runtime:
                kwargs['type'] = runtime.split('-', 1)[1]
            rt = SeleniumRuntime(**kwargs)
            rt.launch_tab(url)
            launched = True
        
        elif runtime.endswith('-browser'):
            # Browser runtime
            runtime = runtime.split('-')[0]
            
            # Try using our own runtimes to open in tab, because
            # the webbrowser module is not that good at opening specific browsers.
            Runtime = _runtimes.get(runtime, None)
            if Runtime is not None:
                rt = Runtime(**kwargs)
                if rt.is_available():
                    rt.launch_tab(url)
                    launched = True
            
            # Use browser runtime (i.e. webbrowser module)
            # Default-browser always works (from the runtime perspective)
            kwargs['type'] = runtime
            rt = BrowserRuntime(**kwargs)
            if rt.is_available():
                rt.launch_tab(url)
                launched = True
        else:
            logger.warn('Runtime names should be "app", "browser" or '
                        'end with "-app" or "-browser", not %r' % runtime)
    
    except Exception as err:
        type_, value, tb = sys.exc_info()
        trace = traceback.format_list(traceback.extract_tb(tb))
        del tb
        return rt, False, str(err) + '\n' + ''.join(trace[-1:])
    
    return rt, launched, None


def _expand_runtime_name(runtime):
    """ Apply aliases and map "x or y" to ['x', 'y'].
    """
    # Normalize
    for c in (' or ', ',', '|'):
        runtime = runtime.replace(c, ' ')
    # Expand
    runtimes = []
    for runtime in runtime.split(' '):
        runtime = runtime.strip().lower()
        if runtime in _aliases:
            runtimes.extend(_expand_runtime_name(_aliases[runtime]))
        else:
            runtimes.append(runtime)
    # Deduplicate
    runtimes2 = []
    for runtime in runtimes:
        if runtime not in runtimes2:
            runtimes2.append(runtime)
    return runtimes2


##

def _print_autoclasses():  # pragma: no cover
    """ Run this code to get ``.. autoclass::`` definitions to put in the docs.
    """
    from webruntime import _runtimes
    lines = []
    seen = set()
    for name in ('BaseRuntime', 'DesktopRuntime'):
        lines.append('.. autoclass:: webruntime.%s\n  :members:' % name)
    for name, Cls in _runtimes.items():
        if Cls in seen:
            continue
        seen.add(Cls)
        lines.append('.. autoclass:: webruntime.%s\n  :members:' % Cls.__name__)
    print('\n\n'.join(lines))



"""
Verify that the app runtimes do not leave profile directories around. We've
had firefox-app spam the  user directory before, and we want to make sure that does not
happen again.
"""

import os
import time
import tempfile

import webruntime
from webruntime.util.testing import run_tests_if_main, skip

userdir = os.path.expanduser('~')


def index():
    """ Get a set of filenames for the current user directory.
    """
    filenames = set()
    for root, dirs, files in os.walk(userdir):
        for fname in dirs:
            # if 'temp' in fname.lower():
            #     continue  # exception for FF, which seems to clean up itself
            filenames.add(os.path.join(root, fname))
        for fname in files:
            if len(fname) < 15:  # only simple names, not uids
                if 'fontconfig' in root:
                    continue  # CI
            filenames.add(os.path.join(root, fname))
    return filenames


def run_once(html_filename, runtime):
    x = webruntime.launch(html_filename, runtime)
    time.sleep(1.5)
    x.close()
    time.sleep(0.5)


def notrailtester(runtime, n=4):
    
    html_filename = os.path.join(tempfile.gettempdir(), 'webruntime_empty_page.html')
    with open(html_filename, 'wb') as f:
        f.write('<html><body>test page</body></html>'.encode())
    
    # Apparently, CI needs some more time for FF init its file system
    # It seems that even ~/Desktop is not there initially, which we use to
    # detect when we're good to go.
    desktop = os.path.normpath(os.path.expanduser('~/Desktop'))
    for iter in range(3):
        run_once(html_filename, runtime)
        before = index()
        if desktop in before:
            break
        
        etime = time.time() + 5  # dont get stuck
        while time.time() < etime and desktop not in before:
            time.sleep(0.2)
            before = index()
    
    print('has desktop:', desktop in before, iter)
    
    # Get all current files in home dir
    before = index()
    
    # Run an app a few times
    for i in range(n):
        run_once(html_filename, runtime)
    
    # Get files that we have now
    after = index()
    
    # Analyze
    extra_files = after.difference(before)
    extra_files2 = [f for f in extra_files
                    if not f.startswith((webruntime.TEMP_APP_DIR,
                                         webruntime.RUNTIME_DIR))]
    
    print(extra_files2)
    assert len(extra_files2) < n


def test_notrail_firefox():
    if not webruntime.FirefoxRuntime().is_available():
        skip('no firefox')
    notrailtester('firefox-app')


def test_notrail_nw():
    if not webruntime.NWRuntime().is_available():
        skip('no nw')
    notrailtester('nw-app')


if __name__ == '__main__':
    test_notrail_firefox()
    test_notrail_nw()
    # run_tests_if_main()



""" Web runtime based on NW.js

https://github.com/nwjs/nw.js


Dev notes
---------

Apps must have a way to be "unique", so that if you create 2 apps with Flexx,
they won't group on the taskbar, and are preferably easily discovered in the
process manager.

The "name" in the manifest is one option. Note that an appdata dir is
created for each used name, unless one sets --user-data-dir, as we do.

The "description" in the manifest, not sure where that turns up.

The name of the executable is the main thing to change.

"""

import os.path as op
import os
import sys
import json
import tempfile

from . import config
from ._common import DesktopRuntime
from ._manage import create_temp_app_dir
from ._manage import open_arch, extract_arch, versionstring


# http://docs.nwjs.io/en/latest/References/Manifest%20Format
def get_manifest_template():
    return {"name": "webruntime_nw_app",
            "main": "",
            "nodejs": False,
            # "single-instance": False,  # Deprecated; is always True
            "description": "an app launched with webruntime",
            "version": "1.0",
            "keywords": [],
            
            "window": {
                "title": "",
                "icon": "",
                "toolbar": False,
                "frame": True,
                "width": 640,
                "height": 480,
                "position": "center",
                "resizable": True,
                "min_width": 10,
                "min_height": 10,
                #"max_width": 800,
                #"max_height": 600,
                "always-on-top": False,
                "fullscreen": False,
                "kiosk": False,
                "transparent": False,
                "show_in_taskbar": True,
                "show": True
                },
            
            "webkit": {
                "plugin": True,
                "java": False
                },
            
            # Tweak chromium: es3 == webgl2, single process makes things
            # more compact, but does not work on OS X.
            "chromium-args": ' '.join(['--enable-unsafe-es3-apis',
                                       # '--single-process',
                                       ]),
            }


def fix_libudef(dest):
    """ Fix the dependency for libudef by making a link to libudef.so.1.
    
    github.com/rogerwang/node-webkit/wiki/The-solution-of-lacking-libudev.so.0 
    """
    
    paths = ["/lib/x86_64-linux-gnu/libudev.so.1",  # Ubuntu, Xubuntu, Mint
             "/usr/lib64/libudev.so.1",  # SUSE, Fedora
             "/usr/lib/libudev.so.1",  # Arch, Fedora 32bit
             "/lib/i386-linux-gnu/libudev.so.1",  # Ubuntu 32bit
             ]
    
    target = op.join(dest, 'libudev.so.0')
    for path in paths:
        if op.isfile(path) and not op.isfile(target):
            os.symlink(path, target)


class NWRuntime(DesktopRuntime):
    """ Desktop runtime based on NW.js (http://nwjs.io, formerly
    node-webkit), which uses the Chromium engine. We will install/update the
    runtime if it finds a suitable archive in a few common locations like
    the desktop, download dir and temp dir. That way, the end-user only
    has to download the archive to make this runtime work.
    
    This runtime is visible in the task manager under a custom process name
    (``sys.executable + '-ui'``), making it easy to spot in the task manager,
    and avoids task-bar grouping. Compared to the Firefox app runtime,
    this runtime uses more processes and memory. Chrome used to be faster, but
    since Firefox Quantum this is no longer the case.
    
    The supported ``windowmode`` values are 'normal', 'fullscreen',
    'kiosk' ('maximized' is ignored).
    """
    
    _pending_install = None
    
    def _get_name(self):
        return 'nw'
    
    def _get_install_instuctions(self):
        m = ('Download the NW.js archive for your platform from '
             '"http://nwjs.io". The webruntime library will find the file if '
             'it is placed in your home dir, desktop, downloads dir (where '
             'most browser save it) or the default temp dir.')
        return m
    
    def _get_exe_name(self, dir):
        if sys.platform.startswith('win'):
            return op.join(dir, 'nw.exe')
        elif sys.platform.startswith('darwin'):
            return op.join(dir, 'nwjs.app', 'Contents', 'MacOS', 'nwjs')
        else:
            return op.join(dir, 'nw')
    
    def _get_exe(self):
        # Get exe to up-to-date locally installed
        # (this does an install/update if necessary)
        exe = op.realpath(self._get_exe_name(self.get_runtime_dir()))
        if op.isfile(exe):
            return exe
    
    def _install_runtime(self, archive, path):
        extract_arch(open_arch(archive), path)
    
    def _get_system_version(self):
        """ Get the version and filename of any valid nwjs archive for this
        platform.
        """
        
        # Where can we look for archives
        dirs = [op.expanduser('~'),
                op.expanduser('~/Desktop'),
                op.expanduser('~/Downloads'),
                tempfile.gettempdir(),
                ]
        
        # What are we looking for?
        exts = '.zip', '.tar', '.tar.gz', '.tar.bz2'
        if sys.platform.startswith('win'):
            plat = '-win'
        elif sys.platform.startswith('linux'):
            plat = '-linux'
        elif sys.platform.startswith('darwin'):
            plat = '-osx'
        
        # Collect archives
        archives = {}  # version -> filename
        for dir in dirs:
            if os.path.isdir(dir):
                for fname in os.listdir(dir):
                    if fname.startswith('nwjs-v') and plat in fname:
                        if fname.lower().endswith(exts):
                            version = fname.split('-v')[1].split('-')[0]
                            archives[version] = os.path.join(dir, fname)
        
        # Avoid having to open archives which we know are not of higher version
        version_th, _ = self.get_cached_version()
        
        # Try - highest version first - whether the archive is ok
        for version in reversed(sorted(archives, key=versionstring)):
            if version_th and versionstring(version) <= versionstring(version_th):
                break
            try:
                with open_arch(archives[version]):
                    pass
            except Exception:
                continue
            else:
                return version, archives[version]
        
        return None, None
    
    def _get_version(self):
        self.get_exe()
        cur_version, _ = self.get_cached_version()
        return cur_version
    
    def _test_platform(self):
        if not (sys.platform.startswith('win') or
                sys.platform.startswith('linux') or
                sys.platform.startswith('darwin')):
            raise RuntimeError('NW.js runtime is not supported on platform ' +
                               sys.platform)
    
    def _launch_tab(self, url):
        raise RuntimeError('NW runtime cannot launch tabs.')
    
    def _launch_app(self, url):
        
        self._test_platform()
        self._clean_nw_dirs()
        
        # Get dir to store app definition
        app_path = create_temp_app_dir('nw')
        id = op.basename(app_path).split('_', 1)[1].replace('~', '_')
        
        # Get runtime exe
        if config.nw_exe:
            # User specifies the executable, we're not going to worry about version
            exe = config.nw_exe
        else:
            # We install the runtime, based on a minimal required version
            exe = self.get_exe()
            
            # Change exe to avoid grouping + easier recognition in task manager
            if exe and op.isfile(op.realpath(exe)):
                exe = self._get_app_exe(exe, app_path)
        
        # Populate app definition
        # name must be a unique, lowercase alpha-numeric name without spaces.
        # It may include "." or "_" or "-" characters. Normally, NW.js stores
        # the app's profile data under the directory named name, but we
        # overload user-data-dir.
        
        # From 0.20.0, even with --user-data-dir, a profile dir with "name" is
        # still created (at least on Windows). Fortunately. the name does not
        # have to be unique, perhaps because we define a custom profile dir.
        D = get_manifest_template()
        D['name'] = 'webruntime_stub_nw_profile'
        D['description'] += ' (%s)' % id
        D['main'] = url
        D['window']['title'] = self._title
        
        # Set size (position can be null, center, mouse)
        D['window']['kiosk'] = self._windowmode == 'kiosk'
        D['window']['fullscreen'] = self._windowmode == 'fullscreen'
        D['window']['width'], D['window']['height'] = self._size
        
        # Icon (note that icon is "overloaded" if nw is wrapped in a runtime.app)
        self._icon.write(op.join(app_path, 'app.png'))  # ico does not work
        size = [i for i in self._icon.image_sizes() if i <= 64][-1]
        D['window']['icon'] = 'app%i.png' % size
        
        # Write app manifest
        with open(op.join(app_path, 'package.json'), 'wb') as f:
            f.write(json.dumps(D, indent=4).encode())
        
        # Fix libudef bug
        llp = os.getenv('LD_LIBRARY_PATH', '')
        if sys.platform.startswith('linux'):
            fix_libudef(app_path)
            llp = app_path + os.pathsep + llp
        
        # Prepare profile dir for NW/Chromium to let --user-data-dir point to.
        # This dir is unique for each instance of the app, but because it is
        # inside the app_path, it gets automatically cleaned up.
        profile_dir = op.join(app_path, 'stub_profile')
        if not op.isdir(profile_dir):
            os.mkdir(profile_dir)
        
        # Launch
        cmd = [exe, '--user-data-dir=' + profile_dir, app_path]
        self._start_subprocess(cmd, LD_LIBRARY_PATH=llp)
    
    def _clean_nw_dirs(self):
        """ NW makes empty dirs in temp dir, clean these up.
        """
        if sys.platform.startswith('win'):  # only an issue on Windows
            dir = tempfile.gettempdir()
            for dname in os.listdir(dir):
                if dname.startswith('nw'):
                    dirname = os.path.join(dir, dname)
                    if not os.listdir(dirname):
                        try:
                            os.rmdir(dirname)
                        except Exception:
                            pass



""" Web runtime based on a chrome/chromium.

It is possible to make a chrome app with a custom icon on Windows (because
it uses the (initial) favicon of the page) and OS X (because how apps work).
I tried hard to make it work on Linux via a .desktop file, but the problem
is that Chome explicitly sets its icon (Chromium does not). Further, both
Chrome and Chromium reset the process name (arg zero), causing the app to be
grouped with Chrome. This makes Chrome not an ideal runtime for apps; use
the NW runtime to effectively make use of the Chromium runtime.
"""

import os.path as op
import os
import sys
import subprocess

from . import config
from ._common import DesktopRuntime, find_osx_exe


class ChromeRuntime(DesktopRuntime):
    """ Runtime representing either the Google Chrome or Chromium browser.
    This runtime does support desktop-like apps, which works pretty well on
    Windows, but not so much on OS X and Linux:

    * On Linux it has the Chrome/Chromium icon.
    * On OSX and Linux it groups with the Chrome/Chromium browser.
    * Fullscreen mode does not work on OS X.
    """
    
    # Note, this is not an abstract class, but a proxy class for either browser
    
    def _get_name(self):
        return 'chrome'
    
    def _get_install_instuctions(self):
        m = 'Install Chrome from http://chrome.com or https://www.chromium.org'
        if sys.platform.startswith('linux'):
            m += ', or use your package manager.'
        return m
    
    def _get_version(self, exe=None):
        if exe is None:
            exe = self.get_exe()
            if exe is None:
                return
        
        # Get raw version string (as bytes)
        if sys.platform.startswith('win'):
            if not op.isfile(exe):
                return
            version = subprocess.check_output(['wmic', 'datafile', 'where',
                                               'name=%r' % exe,
                                               'get', 'Version', '/value'])
        else:
            version = subprocess.check_output([exe, '--version'])
        
        # Clean up
        parts = version.decode(errors='ignore').strip().replace('=', ' ').split(' ')
        for part in parts:
            if part and part[0].isnumeric():
                return part
    
    def _get_system_version(self):
        return self.get_version(), self.get_exe()
    
    def _install_runtime(self, exe, path):
        """ Symlink on Unix. Stub in Windows. In contrast to XUL, exe
        renaming is not needed to avoid grouping on Windows. Plus the
        firewall would asks permission for each new exe that we use
        this way. Ironically, there seems to be nothing we can do to avoid
        grouping on Linux and OS X.
        """
        if sys.platform.startswith('win'):
            os.mkdir(path)
            with open(op.join(path, 'stub.txt'), 'wb') as f:
                f.write('Webruntime uses the system Chrome'.encode())
        else:
            # This makes a nice icon on OS X, but it still groups with Chrome!
            os.mkdir(path)
            os.symlink(exe, op.join(path, 'chrome'))
    
    def _launch_tab(self, url):
        self._spawn_subprocess([self.get_exe(), url])
    
    def _launch_app(self, url):
        
        # Don't bother with create_temp_app_dir(); its only advantage would be
        # having an icon on OS X, but since it's grouped with Chrome, leave it.
        
        # Get chrome executable, but don't really use it
        self.get_runtime_dir()
        
        exe = self.get_exe()
        if exe is None:
            raise RuntimeError('Chrome/Chromium is not available on this system.')
        
        # No way to set icon and title. On Windows, Chrome uses document
        # title/icon. On OS X we create an app. On Linux ... tough luck
        # self._title ...
        # self._icon ...
        
        # Options
        opts = ['--incognito']
        opts.append('--enable-unsafe-es3-apis')  # enable webgl2
        opts.append('--window-size=%i,%i' % self._size)
        if self._pos:
            opts.append('--window-position=%i,%i' % self._pos)
        if self._windowmode == 'kiosk':
            opts.append('--kiosk')
        elif self._windowmode == 'fullscreen':
            opts.append('--start-fullscreen')
        elif self._windowmode == 'maximized':
            opts.append('--start-maximized')
        
        # Launch url, important to put opts before --app=xx
        self._start_subprocess([exe] + opts + ['--app=%s' % url])
        # self._spawn_subprocess([exe] + opts + ['--app=%s' % url])
    
    def _get_exe(self):
        return self._get_google_chrome_exe() or self._get_chromium_exe()
    
    def _get_google_chrome_exe(self):
        # Return user-specified version?
        # Note that its perfectly fine to specify a chromium exe here 
        if config.chrome_exe and self._get_version(config.chrome_exe):
            return config.chrome_exe
        
        paths = []
        
        # Collect possible locations
        if sys.platform.startswith('win'):
            paths.append(r"C:\Program Files\Google\Chrome\Application")
            paths.append(r"C:\Program Files (x86)\Google\Chrome\Application")
            paths.append(r"~\AppData\Local\Google\Chrome\Application")
            paths.append(r"~\AppData\Local\Google\Chrome SxS\Application")
            paths.append(r"~\Local Settings\Application Data\Google\Chrome")
            paths = [op.expanduser(p + '\\chrome.exe') for p in paths]
        elif sys.platform.startswith('linux'):
            paths.append('/usr/bin/chromium')
            paths.append('/usr/bin/chromium-browser')
            paths.append('/snap/bin/chromium)')
            paths.append('/usr/bin/google-chrome')
            paths.append('/usr/bin/google-chrome-stable')
            paths.append('/usr/bin/google-chrome-beta')
            paths.append('/usr/bin/google-chrome-dev')
        elif sys.platform.startswith('darwin'):
            app_dirs = [
                '~/Applications/Chrome',
                '~/Applications/Google Chrome',
                '/Applications/Chrome',
                '/Applications/Google Chrome',
                '/Applications/Google Chrome.app',
                '/Applications/Chromium.app',
                '/Applications/Google Chrome Canary.app',
                find_osx_exe('com.google.Chrome')
            ]
            for dir in app_dirs:
                if dir:
                    dir = op.expanduser(dir)
                    if op.isdir(dir):
                        paths.append(op.join(dir, 'Contents/MacOS/Chrome'))
                        paths.append(op.join(dir, 'Contents/MacOS/Google Chrome'))
                        paths.append(op.join(dir, 'Contents/MacOS/Chromium'))
                        paths.append(
                            op.join(dir, 'Contents/MacOS/Google Chrome Canary')
                        )
        
        # Try location until we find one that exists
        for path in paths:
            if op.isfile(path):
                return path
        
        # Getting desperate ...
        for path in os.getenv('PATH', '').split(os.pathsep):
            if 'chrome' in path.lower():
                for name in ('chrome.exe', 'chrome', 'google-chrome', 'Google Chrome'):
                    if op.isfile(op.join(path, name)):
                        return op.join(path, name)
        
        # We cannot find it
        return None

    
    def _get_chromium_exe(self):
        paths = []
        
        # Collect possible locations
        if sys.platform.startswith('win'):
            paths.append(r"C:\Program Files\Chromium\Application")
            paths.append(r"C:\Program Files (x86)\Chromium\Application")
            paths.append(r"~\AppData\Local\Chromium\chrome.exe")
            paths.append(r"~\AppData\Local\Chromium\Application")
            paths.append(r"~\Local Settings\Application Data\Chromium")
            paths = [op.expanduser(p + '\\chrome.exe') for p in paths]
        elif sys.platform.startswith('linux'):
            paths.append('/usr/bin/chromium')
            paths.append('/usr/bin/chromium-browser')
        elif sys.platform.startswith('darwin'):
            app_dirs = ['~/Applications/Chromium', '~/Applications/Chromium',
                        find_osx_exe('org.chromium.Chromium')]
            for dir in app_dirs:
                if dir:
                    dir = op.expanduser(dir)
                    if op.isdir(dir):
                        paths.append(op.join(dir, 'Contents/MacOS/Chromium'))
                        paths.append(op.join(dir, 'Contents/MacOS/Chromium Browser'))
        
        # Try location until we find one that exists
        for path in paths:
            if op.isfile(path):
                return path
        
        # Getting desperate ...
        for path in os.getenv('PATH', '').split(os.pathsep):
            if 'chromium' in path.lower():
                for name in ('chrome.exe', 'chromium',
                             'chromium-browser', 'Chromium Browser'):
                    if op.isfile(op.join(path, name)):
                        return op.join(path, name)
        
        # We cannot find it
        return None


class GoogleChromeRuntime(ChromeRuntime):
    """ Runtime based on the Google Chrome browser. Derives from ChromeRuntime.
    """
    
    def _get_name(self):
        return 'googlechrome'
    
    def _get_exe(self):
        return self._get_google_chrome_exe()
    
    def _get_install_instuctions(self):
        return 'Install Google Chrome from http://chrome.com'


class ChromiumRuntime(ChromeRuntime):
    """ Runtime based on the Chromium browser. Derives from ChromeRuntime.
    """
    
    def _get_name(self):
        return 'chromium'
    
    def _get_exe(self):
        return self._get_chromium_exe()
    
    def _get_install_instuctions(self):
        m = 'Install Chromium from https://www.chromium.org'
        if sys.platform.startswith('linux'):
            m += ', or use your package manager.'
        return m



from .util.config import Config

config = Config('webruntime', '~appdata/.flexx.cfg', '~appdata/.webruntime.cfg',

        # webruntime=('', str, 'The default web runtime to use. '
        #             'Default is "app or browser".'),
        firefox_exe=('', str, 'The location of the Firefox executable. '
                     'Auto-detect by default.'),
        chrome_exe=('', str, 'The location of the Chrome/Chromium executable. '
                    'Auto-detect by default.'),
        nw_exe=('', str, 'The location of the NW.js executable. '
                'Auto-install by default.'),
        )



""" Web runtime based on qt-webkit. Requires PyQt4 or PySide.
"""

import os
import sys

from ._common import DesktopRuntime
from ._manage import create_temp_app_dir

# Note that setting icon on Ubuntu (and possibly on other OS-es is broken for PyQt)

# Note that this runtime could allow us a very high degree of control
# by e.g. passing commands through stdin. However, Qt webkit cannot
# render Flexx's ui apps.

CODE_TO_RUN = """
import sys

qt = None
for iter in range(4):
    try:
        if iter == 0:
            from PyQt5 import QtCore, QtGui, QtWebKit, QtWidgets, QtWebKitWidgets
            qt = 'pyqt5'
        elif iter == 1:
            from PyQt4 import QtCore, QtGui, QtWebKit
            qt = 'pyqt4'
        elif iter == 2:
            from PySide import QtCore, QtGui, QtWebKit
            qt = 'pyside'
        elif iter == 3:
            from PySide2 import QtCore, QtGui, QtWebKit, QtWidgets, QtWebKitWidgets
            qt = 'pyside2'
        break
    except ImportError:
        pass

if not qt:
    sys.exit('Cannot import Qt')

url = {url}
title = {title}
icon = {icon}
size = {size}
pos = {pos}

if qt not in ['pyqt5', 'pyside2']:
    app = QtGui.QApplication([])
    m = QtWebKit.QWebView(None)
    
else:
    app = QtWidgets.QApplication([])
    m = QtWebKitWidgets.QWebView(None)

m.setUrl(QtCore.QUrl(url))
m.setWindowTitle(title + ' (%s)' % qt)
if icon:
    i = QtGui.QIcon()
    i.addFile(icon, QtCore.QSize(16, 16))
    m.setWindowIcon(i)
if size:
    m.resize(*size)
if pos:
    m.move(*pos)

m.show()
app.exec_()
"""

# todo: maximized / fullscreen can certainly be implemented,
# but I don't care much about this runtime now, so I did not do it yet.


class PyQtRuntime(DesktopRuntime):
    """ Desktop runtime based on qt-webkit. Launches a new Python
    process (the same version as the current), and uses PyQt4, PyQt5 or PySide
    to display the page.
    
    This runtime is not suited for hosting apps created with e.g. flexx.ui;
    it is included for completeness but should generally be avoided.
    """
    
    def _get_name(self):
        return 'pyqt'
    
    def _get_install_instuctions(self):
        return ('To enable the Pyqt runtime, install Pyqt5, Pyqt4, '
                'Pyside2 or Pyside in your Python environment.')
    
    def _get_exe(self):
        return sys.executable
        # todo: perhaps we should test whether pyqt is available, but
        # don't want to do that by importing it... meeh, this runtime
        # is not that serious anyway.
    
    def _get_version(self):
        return None  # we could report the Qt version here (or pyqt version?)
    
    def _get_system_version(self):
        return None, None  # stub
    
    def _install_runtime(self, sys_path, dest_path):
        pass  # is never called
    
    def _launch_tab(self, url):
        raise RuntimeError('PyQt runtime cannot launch tabs.')
    
    def _launch_app(self, url):
        
        # We don't call self.get_runtime_dir() so we don't need to implement
        # _install_runtime()
        
        # Write icon - assume that we have a 16x16 icon
        app_path = create_temp_app_dir('pyqt')
        self._icon.write(os.path.join(app_path, 'icon.png'))
        iconfile = os.path.join(app_path, 'icon%i.png' % self._icon.image_sizes()[0])
        
        code = CODE_TO_RUN.format(url=repr(url),
                                  title=repr(self._title),
                                  icon=repr(iconfile),
                                  size=repr(self._size),
                                  pos=repr(self._pos),
                                  )
        self._start_subprocess([self.get_exe(), '-c', code])



from __future__ import print_function, absolute_import

import re
import os
import sys
import logging

if sys.version_info[0] == 2:  # pragma: no cover
    from ConfigParser import ConfigParser as _ConfigParser
    from StringIO import StringIO

    class ConfigParser(_ConfigParser):
        def read_string(self, string, source):
            return self.readfp(StringIO(string), source)

else:
    from configparser import ConfigParser


def as_bool(value):
    if isinstance(value, bool):
        return value
    elif isinstance(value, basestring) and value.lower() in BOOLEAN_STATES:
        return BOOLEAN_STATES[value.lower()]
    else:
        raise ValueError('Cannot make a bool of %r' % value)

def get_tuple_validator(subvalidator):
    def validator(value):
        if isinstance(value, (tuple, list)):
            value2 = tuple(value)
        elif isinstance(value, basestring):
            value2 = tuple([s.strip() for s in value.strip('()[]').split(',')])
        else:
            raise ValueError('Cannot make a tuple of %r' % value)
        return tuple([subvalidator(x) for x in value2])
    return validator

def stack_sorter(key):
    # Implement ordering, files and strings go at spot 1
    return dict(default=0, environ=2, argv=3, set=4).get(key[0], 1)


BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                  '0': False, 'no': False, 'false': False, 'off': False}

TYPEMAP = {float: float, int: int, bool: as_bool}
if sys.version_info[0] == 2:  # pragma: no cover
    TYPEMAP[basestring] = unicode  # noqa
    TYPEMAP[str] = unicode  # noqa
else:
    basestring = str
    TYPEMAP[str] = str

INSTANCE_DOCS = """ Configuration object for {name}

    The options below can be set from different sources, and are
    evaluated in the following order:

    * From the default value.
    * From .cfg or .ini file, or a string in cfg format.
    * From environment variables, e.g. ``{NAME}_FOO=3``.
    * From command-line arguments, e.g. ``--{name}-foo=3``.
    * From setting the config option directly, e.g. ``config.foo = 3``.

    Use ``print(config)`` to get a summary of the current values and
    from which sources they were set.

    Parameters:
    """


class Config(object):
    """ Class for configuration objects.

    A Config object has a set of options, which can be str, int, float,
    bool, or a tuple of any of the above. Options can be set from
    different sources:

    * Each option has a default value.
    * From .cfg or .ini files.
    * From strings in ini format.
    * From environment variables.
    * From command-line arguments.
    * By setting the config option directly.

    Parameters:
        name (str): the name by which to identify this config. This name
            is used as a prefix in environment variables and command
            line arguments, and optionally as a section header in .cfg files.
        *sources: Sources to initialize the option values with.
            These can be strings in ini format, or .ini or .cfg filenames.
            If a file is given that does not exist, it is simply ignored.
            Special prefixes ``~/`` and ``~appdata/`` are expanded to the
            home dir and appdata dir.
        **options: The options specification: each option consists of
            a 3-element tuple (default, type, docstring).

    Example:

        .. code-block:: Python

            config = Config('myconfig', '~appdata/.myconfig.cfg',
                            foo=(False, bool, 'Whether to foo'),
                            bar=(0.0, float, 'The size of the bar'),
                            spam=('1,2,3', [int], 'A tuple of ints'))

        With this, options can be set:

        * With an entry ``foo = 3`` in "~appdata/.myconfig.cfg".
        * With a string ``"foo = 3"`` passed at initialization.
        * With an environment variable named ``MYCONFIG_FOO``.
        * With a command line argument ``--myconfig-foo=3``.
        * By doing ``config.foo = 3``, or ``config['foo'] = 3`` in Python.

    Notes:
        * Option names are case insensitive, except for attribute access and
          environment variables (the latter must be all uppercase).
        * All values can be set as a Python object or a string; they
          are automatically converted to the correct type.
        * Each instance gets a docstring that lists all options, so it
          can easily be used in e.g. Sphynx docs.
    """

    def __init__(self, name, *sources, **options):

        # The identifier name for this config
        self._name = name
        if not is_valid_name(name):
            raise ValueError('Config name must be an alphanumeric string, '
                             'starting with a letter.')

        # The option names (unmodified case)
        self._options = []

        # Where the values are stored, we keep a stack, lowercase keys
        self._opt_values = {}  # name -> list of (source, value) tuples

        # Map of lowercase option names to validator functions
        self._opt_validators = {}

        # Map of lowercase option names to type names, for better reporting
        self._opt_typenames = {}

        # Map of lowercase option names to docstrings
        self._opt_docs = {}

        # Parse options
        option_docs = ['']
        for name in sorted(options.keys(), key=lambda x: x.lower()):
            lname = name.lower()
            spec = options[name]
            # Checks
            if not is_valid_name(name):
                raise ValueError('Option name must be alphanumeric strings, '
                                 'starting with a letter, and not private.')
            if not len(spec) == 3:
                raise ValueError('Option spec must be (default, type, docs)')
            default, typ, doc = spec
            istuple = False
            if isinstance(typ, (tuple, list)):
                if len(typ) != 1:
                    raise ValueError('Tuple type spec should have one element.')
                istuple, typ = True, typ[0]
            if not (isinstance(typ, type) and issubclass(typ, tuple(TYPEMAP))):
                raise ValueError('Option types can be str, bool, int, float.')
            # Parse
            typename = typ.__name__ + ('-tuple' if istuple else '')
            args = name, typename, doc, default
            option_docs.append(' '*8 + '%s (%s): %s (default %r)' % args)
            self._options.append(name)
            self._opt_typenames[lname] = typename
            self._opt_validators[lname] = (get_tuple_validator(TYPEMAP[typ])
                                           if istuple else TYPEMAP[typ])
            self._opt_docs[lname] = doc
            self._opt_values[lname] = []

        # Overwrite docstring
        self.__doc__ = INSTANCE_DOCS.format(name=self._name,
                                            NAME=self._name.upper())
        self.__doc__ += '\n'.join(option_docs)

        # --- init values

        # Set defaults
        for name, spec in options.items():
            self._set('default', name, spec[0])

        # Load from sources
        for source in sources:
            if not isinstance(source, basestring):
                raise ValueError('Sources should be strings or filenames.')
            if '\n' in source:
                self.load_from_string(source)
            else:
                self.load_from_file(source)

        # Load from environ
        for name in self._opt_values:
            env_name = (self._name + '_' + name).upper()
            value = os.getenv(env_name, None)  # getenv is case insensitive
            if value is not None:
                self._set('environ', name, value)

        # Load from argv
        arg_prefix = '--' + self._name.lower() + '-'
        for i in range(1, len(sys.argv)):
            arg = sys.argv[i]
            if arg.lower().startswith(arg_prefix) and '=' in arg:
                name, value = arg[len(arg_prefix):].split('=', 1)
                if name.lower() in self._opt_values:
                    self._set('argv', name, value)

    def __repr__(self):
        t = '<Config %r with %i options at 0x%x>'
        return t % (self._name, len(self._options), id(self))

    def __str__(self):
        # Return a string representing a summary of the options and
        # how they were set from different sources.
        lines = []
        lines.append('Config %r with %i options.' %
                     (self._name, len(self._options)))
        for name in self._options:
            lname = name.lower()
            lines.append('\nOption %s (%s) - %s' % (name,
                                                  self._opt_typenames[lname],
                                                  self._opt_docs[lname]))
            for source, val in self._opt_values[lname]:
                lines.append('    %r from %s' % (val, source))
            lines[-1] = ' -> ' + lines[-1][4:]  # Mark current value
        return '\n'.join(lines)

    def __len__(self):
        return len(self._options)

    def __iter__(self):
        return self._options.__iter__()

    def __dir__(self):
        return self._options

    def __getattr__(self, name):
        # Case sensitive get
        if not name.startswith('_') and name in self._options:
            return self._opt_values[name.lower()][-1][1]
        return super(Config, self).__getattribute__(name)

    def __getitem__(self, name):
        # Case insensitive get
        if not isinstance(name, basestring):
            raise TypeError('Config only allows subscripting by name strings.')
        if name.lower() in self._opt_values:
            return self._opt_values[name.lower()][-1][1]
        else:
            raise IndexError('Config has no option %r' % name)

    def __setattr__(self, name, value):
        # Case sensitive set
        if not name.startswith('_') and name in self._options:
            return self._set('set', name, value)
        return super(Config, self).__setattr__(name, value)

    def __setitem__(self, name, value):
        # Case insensitve set
        if not isinstance(name, basestring):
            raise TypeError('Config only allows subscripting by name strings.')
        if name.lower() in self._opt_values:
            return self._set('set', name, value)
        else:
            raise IndexError('Config has no option %r' % name)

    def _set(self, source, name, value):
        # The actual setter (case insensitive), applies the validator
        validator = self._opt_validators[name.lower()]
        try:
            real_value = validator(value)
        except Exception:
            args = name, self._opt_typenames[name.lower()], value
            raise ValueError('Cannot set option %s (%s) from %r' % args)
        stack = self._opt_values[name.lower()]
        if stack and stack[-1][0] == source:
            stack[-1] = source, real_value
        else:
            stack.append((source, real_value))
            stack.sort(key=stack_sorter)

    def load_from_file(self, filename):
        """ Load config options from a file, as if it was given as a
        source during initialization. This means that options set via
        argv, environ or directly will not be influenced.
        """
        # Expand special prefix
        filename = filename.replace('~appdata/', appdata_dir() + '/')
        filename = filename.replace('~appdata\\', appdata_dir() + '\\')
        filename = os.path.expanduser(filename)
        # Proceed if is an actual file
        if os.path.isfile(filename):
            text = None
            try:
                text = open(filename, 'rb').read().decode()
            except Exception as err:
                logging.warn('Could not read config from %r:\n%s' %
                             (filename, str(err)))
                return
            self.load_from_string(text, filename)

    def load_from_string(self, text, filename='<string>'):
        """ Load config options from a string, as if it was given as a
        source during initialization. This means that options set via
        argv, environ or directly will not be influenced.
        """
        try:
            self._load_from_string(text, filename)
        except Exception as err:
            logging.warn(str(err))

    def _load_from_string(self, s, filename):
        # Create default section, so that users can work with sectionless
        # files (as is common in an .ini file)
        name_section = '[%s]\n' % self._name
        if name_section not in s:
            s = name_section + s
        s += '\n'
        parser = ConfigParser()
        parser.read_string(s, filename)
        if parser.has_section(self._name):
            for name in self._options:
                if parser.has_option(self._name, name):
                    value = parser.get(self._name, name)
                    self._set(filename, name, value)


def is_valid_name(n):
    return isidentifier(n) and not n.startswith('_')


def isidentifier(s):
    # http://stackoverflow.com/questions/2544972/
    if not isinstance(s, basestring):  # noqa
        return False
    return re.match(r'^\w+$', s, re.UNICODE) and re.match(r'^[0-9]', s) is None


# From pyzolib/paths.py (https://bitbucket.org/pyzo/pyzolib/src/tip/paths.py)
def appdata_dir(appname=None, roaming=False):
    """ Get the path to the application directory, where applications
    are allowed to write user specific files (e.g. configurations).
    """
    # Define default user directory
    userDir = os.path.expanduser('~')
    # Get system app data dir
    path = None
    if sys.platform.startswith('win'):
        path1, path2 = os.getenv('LOCALAPPDATA'), os.getenv('APPDATA')
        path = (path2 or path1) if roaming else (path1 or path2)
    elif sys.platform.startswith('darwin'):
        path = os.path.join(userDir, 'Library', 'Application Support')
    # On Linux and as fallback
    if not (path and os.path.isdir(path)):  # pragma: no cover
        path = os.environ.get(
            "XDG_CONFIG_HOME",
            os.path.expanduser(os.path.join("~", ".config")))
    # Maybe we should store things local to the executable (in case of a
    # portable distro or a frozen application that wants to be portable)
    prefix = sys.prefix
    if getattr(sys, 'frozen', None):  # See application_dir() function
        prefix = os.path.abspath(os.path.dirname(sys.executable))
    for reldir in ('settings', '../settings'):
        localpath = os.path.abspath(os.path.join(prefix, reldir))
        if os.path.isdir(localpath):  # pragma: no cover
            try:
                open(os.path.join(localpath, 'test.write'), 'wb').close()
                os.remove(os.path.join(localpath, 'test.write'))
            except IOError:
                pass  # We cannot write in this directory
            else:
                path = localpath
                break
    # Get path specific for this app
    if appname:  # pragma: no cover
        if path == userDir:
            appname = '.' + appname.lstrip('.')  # Make it a hidden directory
        path = os.path.join(path, appname)
        if not os.path.isdir(path):  # pragma: no cover
            os.mkdir(path)
    # Done
    return path


if __name__ == '__main__':

    sys.argv.append('--test-foo=8')
    c = Config('test',
               foo=(3, int, 'foo yeah'),
               spam=(2.1, float, 'a float!'))



"""
Config and definitions specific to Webruntime.
"""

import os.path as op

from . import ROOT_DIR, THIS_DIR  # noqa

NAME = 'webruntime'
DOC_DIR = op.join(ROOT_DIR, 'docs')
DOC_BUILD_DIR = op.join(DOC_DIR, '_build')



""" Web runtime based on MSHTML, i.e. Microsofts Trident engine.
"""

import os.path as op
import os
import sys

from ._common import BaseRuntime


class MicrosoftRuntime(BaseRuntime):
    """ Base class for IE and Edge runtimes.
    """
    
    def _get_install_instuctions(self):
        avail = 'available' if sys.platform.startswith('win') else 'only available'
        name = self.get_name()
        win = dict(ie='Windows', edge='modern versions of Windows')[name]
        return 'Runtime %s is %s on %s.' % (name, avail, win)
    
    def _get_version(self):
        return None


class IERuntime(MicrosoftRuntime):
    """ Runtime based on IE (Internet Explorer), only for launching in
    the browser.
    """
    
    def _get_name(self):
        return 'ie'
    
    def _get_exe(self):
        if not sys.platform.startswith('win'):
            return None
        
        # Collect possible locations
        paths = []
        for basepath in ('C:\\Program Files\\', 'C:\\Program Files (x86)\\'):
            paths.append(basepath + 'Internet Explorer\\iexplore.exe')
        
        # Try location until we find one that exists
        for path in paths:
            if op.isfile(path):
                return path
        
        # IE not available
        return None
    
    def _launch_tab(self, url):
        self._spawn_subprocess([self.get_exe(), url])

    def _launch_app(self, url):
        raise RuntimeError('IE runtime cannot run as an app.')


class EdgeRuntime(MicrosoftRuntime):
    """ Runtime based on Edge, only for launching in the browser.
    """
    
    def _get_name(self):
        return 'edge'
    
    def _get_exe(self):
        if not sys.platform.startswith('win'):
            return None
        
        path = op.join(os.environ['windir'], 'SystemApps',
                       'Microsoft.MicrosoftEdge_8wekyb3d8bbwe',
                       'MicrosoftEdge.exe')
    
        if op.isfile(path):
            return path
        
        # Edge not available
        return None
    
    def _launch_tab(self, url):
        self._spawn_subprocess(['start', 'microsoft-edge:'+url], shell=True)

    def _launch_app(self, url):
        raise RuntimeError('Edge runtime cannot run as an app.')



# -*- coding: utf-8 -*-

"""
Webruntime setup script.
"""

import os
import sys
import shutil

try:
    import setuptools  # noqa, analysis:ignore
except ImportError:
    pass  # setuptools allows for "develop", but it's not essential

from distutils.core import setup


## Function we need

def get_version_and_doc(filename):
    NS = dict(__version__='', __doc__='')
    docStatus = 0  # Not started, in progress, done
    for line in open(filename, 'rb').read().decode().splitlines():
        if line.startswith('__version__'):
            exec(line.strip(), NS, NS)
        elif line.startswith('"""'):
            if docStatus == 0:
                docStatus = 1
                line = line.lstrip('"')
            elif docStatus == 1:
                docStatus = 2
        if docStatus == 1:
            NS['__doc__'] += line.rstrip() + '\n'
    if not NS['__version__']:
        raise RuntimeError('Could not find __version__')
    return NS['__version__'], NS['__doc__']


def package_tree(pkgroot):
    subdirs = [os.path.relpath(i[0], THIS_DIR).replace(os.path.sep, '.')
               for i in os.walk(os.path.join(THIS_DIR, pkgroot))
               if '__init__.py' in i[2]]
    return subdirs


## Collect info for setup()

THIS_DIR = os.path.dirname(__file__)

# Define name and description
name = 'webruntime'
description = "Launch HTML5 apps in the browser or a desktop-like runtime."

# Get version and docstring (i.e. long description)
version, doc = get_version_and_doc(os.path.join(THIS_DIR, name, '__init__.py'))


## Setup

setup(
    name=name,
    version=version,
    author='Almar Klein and contributors',
    author_email='almar.klein@gmail.com',
    license='(new) BSD',
    url='http://webruntime.readthedocs.io',
    download_url='https://pypi.python.org/pypi/webruntime',
    keywords="GUI, web, runtime, XUL, nwjs",
    description=description,
    long_description=doc,
    platforms='any',
    provides=[name],
    install_requires=['dialite'],
    packages=package_tree(name),
    package_dir={name: name},
    package_data={},
    entry_points={'console_scripts': ['webruntime = webruntime.__main__:main'], },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



"""
Web runtime based on Mozilla's Firefox browser engine

This is a well-developed runtime, with tested good behavior on
Windows, OSX, Linux and Raspberry Pi.

Developer notes
---------------

For this runtime we symlink to the Firefox application on Unix, and copy
the Firefox directory on Windows. That way we can make a symlink/copy of the
executable with a name of our chosing, so that we can avoid taskbar grouping.

"""

import os.path as op
import os
import sys
import time
import shutil
import subprocess

from . import logger, config
from ._common import DesktopRuntime
from ._manage import create_temp_app_dir


## File templates

# The Profile setting makes all apps use the same dummy profile (see issue #150)

APPLICATION_INI = """
[App]
Vendor={vendor}
Name={name}
Version={version}
BuildID={buildid}
ID={id}
Profile=webruntime_xul_stub_profile

[Gecko]
MinVersion=1.8
MaxVersion=200.*
"""

MAIN_XHTML = """
<?xml version="1.0"?>
<?xml-stylesheet href="chrome://global/skin/" type="text/css"?>

<window
    xmlns="http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"
    id="{windowid}"
    title="{title}"
    windowtype="webruntime:main"
    sizemode="{sizemode}"
    >
    <script type="application/javascript"
            src="chrome://{name}/content/main.js" />
    <!-- content or content-primary ? -->
    <browser src="{url}"
             id="content"
             type="content"
             remote="true"
             remoteType="web"
             flex="1"
             disablehistory="true" />
</window>
""".lstrip()

MAIN_JS = """
"""

# todo: persist="screenX screenY width height sizemode"


PREFS_JS = """
// This tells xulrunner what xul file to use
pref("toolkit.defaultChromeURI", "chrome://{name}/content/main.xhtml");

// This line is needed to let window.open work
pref("browser.chromeURL", "chrome://{name}/content/main.xhtml");

// Set features - setting width, height, maximized, etc. here
pref("toolkit.defaultChromeFeatures", "{windowfeatures}");

// debugging prefs, disable these before you deploy your application!
pref("browser.dom.window.dump.enabled", false);
pref("javascript.options.showInConsole", false);
pref("javascript.options.strict", false);
pref("nglayout.debug.disable_xul_cache", false);
pref("nglayout.debug.disable_xul_fastload", false);
"""


# Keep this just in case
def get_firefox_version_unused(exe):
    """Get the version of the given Firefox executable, or None, if the given
    exe does not represent a valid filename.
    """
    version = None
    if exe and op.isfile(exe):
        if 'Contents/MacOS/' in exe:
            inifile = op.join(op.dirname(exe),
                                '../Resources/application.ini')
        else:
            inifile = op.join(op.dirname(exe), 'application.ini')
        for line in open(inifile, 'rb').read().decode().splitlines():
            if line.lower().startswith('version'):
                version = line.split('=')[1].strip()
                break
    return version


class FirefoxRuntime(DesktopRuntime):
    """ Runtime based on Mozilla Firefox. Can be used to open an app in
    Firefox, as well as launch desktop-like apps via Mozilla's XUL framework.
    Available if Firefox is installed.
    
    This runtime is visible in the task manager under a custom process name
    (``sys.executable + '-ui'``), making it easy to spot in the task manager,
    and avoids task-bar grouping. Compared to the NW runtime, this runtime
    is leaner in terms of memory and number of processes, and is also faster.
    """
    
    def _get_name(self):
        return 'firefox'
    
    def _get_install_instuctions(self):
        m = 'Install Mozilla Firefox from http://firefox.com'
        if sys.platform.startswith('linux'):
            m += ', or use your package manager.'
        return m
    
    def _get_exe(self):
        
        # Return user-specified version?
        if config.firefox_exe and self._get_version(config.firefox_exe):
            return config.firefox_exe
        
        paths = []
        
        # Collect possible locations
        if sys.platform.startswith('win'):
            for basepath in ('C:\\Program Files\\', 'C:\\Program Files (x86)\\'):
                paths.append(basepath + 'Mozilla Firefox\\firefox.exe')
                paths.append(basepath + 'Mozilla\\Firefox\\firefox.exe')
                paths.append(basepath + 'Firefox\\firefox.exe')
                paths.append(basepath + 'Firefox Nightly\\firefox.exe')
        elif sys.platform.startswith('linux'):
            paths.append('/usr/lib/firefox/firefox')
            paths.append('/usr/lib64/firefox/firefox')
            paths.append('/usr/lib/iceweasel/iceweasel')
            paths.append('/usr/lib64/iceweasel/iceweasel')
        elif sys.platform.startswith('darwin'):
            osx_user_apps = op.expanduser('~/Applications')
            osx_root_apps = '/Applications'
            paths.append(op.join(osx_user_apps, 'Firefox.app/Contents/MacOS/firefox'))
            paths.append(op.join(osx_root_apps, 'Firefox.app/Contents/MacOS/firefox'))
            if not any([op.isfile(path) for path in paths]):
                # Try harder - use app-id to get the .app path
                try:
                    osx_search_arg='kMDItemCFBundleIdentifier==org.mozilla.firefox'
                    basepath = subprocess.check_output(['mdfind', osx_search_arg])
                    basepath = basepath.rstrip()
                    if basepath:
                        paths.append(op.join(basepath, 'Contents/MacOS/firefox'))
                except (OSError, subprocess.CalledProcessError):
                    pass
    
        # Try location until we find one that exists
        for path in paths:
            if op.isfile(path):
                return path
        
        # Getting desperate ...
        for path in os.getenv('PATH', '').split(os.pathsep):
            if 'firefox' in path.lower() or 'moz' in path.lower():
                for name in ('firefox.exe', 'firefox', 'iceweasel'):
                    if op.isfile(op.join(path, name)):
                        return op.join(path, name)
        
        # Maybe just ... firefox as a command?
        if self._get_version('firefox'):
            return 'firefox'
        
        # We cannot find it
        return None
    
    def _get_version(self, exe=None):
        if exe is None:
            exe = self.get_exe()
            if exe is None:
                return
        
        # Get raw version string (as bytes)
        if sys.platform.startswith('win'):
            if not op.isfile(exe):
                return
            # https://stackoverflow.com/a/4644565/2271927
            version = subprocess.check_output([exe, '--version', '|', 'more'])
            # version = subprocess.check_output(['wmic', 'datafile', 'where',
            #                                    'name=%r' % exe,
            #                                    'get', 'Version', '/value'])
        else:
            version = subprocess.check_output([exe, '--version'])
        
        # Clean up
        parts = version.decode(errors='ignore').strip().replace('=', ' ').split(' ')
        for part in parts:
            if part and part[0].isnumeric():
                return part
    
    def _get_system_version(self):
        return self.get_version(), self.get_exe()
    
    def _install_runtime(self, exe, path):
        """ Make a local "copy" of the firefox runtime. This should put
        a 'xulrunner' executable in the given path.

        On Windows we must make a copy, so that we can make a copy of
        the runtime exe with the name of our choice, to thereby change
        the process name and avoid taskbar grouping with firefox. On
        other systems we use symlinks. On Windows, when symlinks are
        supported (Python 3.2+, Vista+) they hardly ever work because
        the process has no sufficient privileges.

        On OSX we will create an application "X.app", that symlinks all
        files in the firefox runtime. The symlink that we create here
        is only for reference, and does not actually work.
        """
        if sys.platform.startswith('win'):
            # Windows: copy the whole tuntime
            self._copy_xul_runtime(op.dirname(exe), path)
        else:
            # OSX / Linux: create a symlink to xul runtime exe
            os.mkdir(path)
            os.symlink(exe, op.join(path, 'xulrunner'))
    
    def _launch_app(self, url):
        
        self._check_compat()

        # Get dir to store app definition
        app_path = create_temp_app_dir('firefox')
        id = op.basename(app_path).split('_', 1)[1].replace('~', '_')
        
        # Set size and position
        # Maybe interesting window features: alwaysRaised
        windowfeatures = 'resizable=1,minimizable=1,dialog=0,'
        if self._windowmode == 'normal':
            windowfeatures += 'width=%i,height=%i' % self._size
            if self._pos:
                windowfeatures += ',left=%i,top=%i' % self._pos

        # Create files for app
        self._create_xul_app(app_path, id, url, windowfeatures)

        # Get executable for xul runtime (may be None)
        ff_exe = self.get_exe()
        if not ff_exe:
            raise RuntimeError('Firefox is not available on this system.')
        elif not op.isfile(ff_exe):
            # We have no way to wrap things up in a custom app
            exe = ff_exe
        else:
            # We make sure the runtime is "installed" and mangle the name
            xul_exe = op.join(self.get_runtime_dir(), 'xulrunner')
            xul_exe += '.exe' * sys.platform.startswith('win')
            exe = self._get_app_exe(xul_exe, app_path)
        
        # Prepare profile dir for Xul to let -profile dir point to.
        # This dir is unique for each instance of the app, but because it is
        # inside the app_path, it gets automatically cleaned up.
        profile_dir = op.join(app_path, 'stub_profile')
        if not op.isdir(profile_dir):
            os.mkdir(profile_dir)
        
        # Launch
        cmd = [exe, '-app', op.join(app_path, 'application.ini'),
               '-profile', profile_dir]
        self._start_subprocess(cmd)
    
    def _launch_tab(self, url):
        self._spawn_subprocess([self.get_exe(), url])
    
    def _check_compat(self):
        qts = 'PySide', 'PyQt4', 'PyQt5'
        if any([name+'.QtCore' in sys.modules for name in qts]):
            logger.warn("Using the Firefox web runtime and Qt (PySide/PyQt4/PyQt5) "
                        "together may cause problems.")

    def _create_xul_app(self, path, id, url, windowfeatures):
        """ Create the files that determine the XUL app to launch.
        """
        
        modemap = {'kiosk': 'fullscreen'}
        
        # Dict with all values that are injected in the file templates
        D = dict(vendor='Webruntime',
                 name='webruntime_ui_app',
                 version='1.0',
                 buildid='1',
                 id='some.app@webruntime.io',
                 windowid='xx',
                 title=self._title,
                 url=url,
                 sizemode=modemap.get(self._windowmode, self._windowmode),
                 windowfeatures=windowfeatures)
        
        # Create values that need to be unique
        # The name must be unique to avoid all sort of oddities when launching
        # multiple runtimes (as we do in flexx.app tests). Note that we've had
        # problems with the profile dirs being spammed (NW did, now fixed).
        # Also see "Profile=" in APPLICATION_INI above.
        D['name'] = 'webruntime_stub_xul_profile_' + id
        D['windowid'] = 'W' + id
        D['id'] = 'app_' + id + '@webruntime.io'
        
        # Fill in arguments in file contents
        manifest_link = 'manifest chrome/chrome.manifest'
        manifest = 'content {name} content/'.format(**D)
        application_ini = APPLICATION_INI.format(**D)
        main_xhtml = MAIN_XHTML.format(**D)
        main_js = MAIN_JS  # No format (also problematic due to braces)
        prefs_js = PREFS_JS.format(**D)

        # Clear
        if op.isdir(path):
            shutil.rmtree(path)

        # Create directory structure
        for subdir in ('',
                       'chrome', 'chrome/content',
                       'chrome/icons', 'chrome/icons/default',
                       'defaults', 'defaults/preferences',
                       ):
            os.mkdir(op.join(path, subdir))

        # Create files
        for fname, text in [('chrome.manifest', manifest_link),
                            ('chrome/chrome.manifest', manifest),
                            ('application.ini', application_ini),
                            ('defaults/preferences/prefs.js', prefs_js),
                            ('chrome/content/main.js', main_js),
                            ('chrome/content/main.xhtml', main_xhtml),
                            ]:
            with open(op.join(path, fname), 'wb') as f:
                f.write(text.encode())

        # Icon - use Icon class to write a png (Unix) and an ico (Windows)
        # The DesktopRuntime ensures that there always is an icon
        icon_name = op.join(path, 'chrome/icons/default/' + D['windowid'])
        self._icon.write(icon_name + '.ico')
        self._icon.write(icon_name + '.png')
        icon_name = op.join(path, 'chrome/icons/default/default')
        self._icon.write(icon_name + '.ico')
        self._icon.write(icon_name + '.png')
    
    def _copy_xul_runtime(self, dir1, dir2):
        """ Copy the firefox/xulrunner runtime to a new folder, in which
        we rename the firefox exe to xulrunner. This thus creates a xul
        runtime in a location where we have write access. Used to be able
        to set the process name on Windows, and maybe used to distribute
        apps *with* the runtime.
        """
        t0 = time.time()
        # Get extension
        ext = '.exe' if sys.platform.startswith('win') else ''
        # On Rasberry Pi, the xul runtime is in a (linked) subdir
        if op.isdir(op.join(dir1, 'xulrunner')):
            dir1 = op.join(dir1, 'xulrunner')
        # Clear
        if op.isdir(dir2):
            shutil.rmtree(dir2)
        os.mkdir(dir2)
        try:
            # Copy all files except dirs
            for fname in os.listdir(dir1):
                filename1 = op.join(dir1, fname)
                filename2 = op.join(dir2, fname)
                if op.isfile(filename1):
                    shutil.copy2(filename1, filename2)
            # Copy firefox exe -> xulrunner
            for exe_name in ('firefox', 'iceweasel', 'xulrunner', 'firefox'):
                exe = op.join(dir1, exe_name + ext)
                if op.isfile(exe):
                    break
            shutil.copy2(exe, op.join(dir2, 'xulrunner' + ext))
            logger.info('Copied firefox in %1.1f s' % (time.time()-t0))
        except Exception:
            # Clean up
            shutil.rmtree(dir2)
            raise



"""
Code related to managing runtimes and temporary app directories.

We follow a convetion by which files and dirs can be associated with a
certain process by giving it a suffix "~pid". That way, we can detect
when dirs are not in use and have them removed. We also rename files/dirs
that ought to be removed, so that if deletion fails, we can try again later.
"""

import os.path as op
import os
import sys
import time
import stat
import shutil
import tarfile
import zipfile
import subprocess

from . import logger


# From pyzolib/paths.py (https://bitbucket.org/pyzo/pyzolib/src/tip/paths.py)
def appdata_dir(appname=None, roaming=False, macAsLinux=False):
    """ appdata_dir(appname=None, roaming=False,  macAsLinux=False)
    Get the path to the application directory, where applications are allowed
    to write user specific files (e.g. configurations). For non-user specific
    data, consider using common_appdata_dir().
    If appname is given, a subdir is appended (and created if necessary).
    If roaming is True, will prefer a roaming directory (Windows Vista/7).
    If macAsLinux is True, will return the Linux-like location on Mac.
    """

    # Define default user directory
    userDir = op.expanduser('~')

    # Get system app data dir
    path = None
    if sys.platform.startswith('win'):
        path1, path2 = os.getenv('LOCALAPPDATA'), os.getenv('APPDATA')
        path = (path2 or path1) if roaming else (path1 or path2)
    elif sys.platform.startswith('darwin') and not macAsLinux:
        path = op.join(userDir, 'Library', 'Application Support')
    # On Linux and as fallback
    if not (path and op.isdir(path)):
        path = userDir

    # Maybe we should store things local to the executable (in case of a
    # portable distro or a frozen application that wants to be portable)
    prefix = sys.prefix
    if getattr(sys, 'frozen', None):  # See application_dir() function
        prefix = op.abspath(op.dirname(sys.executable))
    for reldir in ('settings', '../settings'):
        localpath = op.abspath(op.join(prefix, reldir))
        if op.isdir(localpath):  # pragma: no cover
            try:
                open(op.join(localpath, 'test.write'), 'wb').close()
                os.remove(op.join(localpath, 'test.write'))
            except IOError:
                pass  # We cannot write in this directory
            else:
                path = localpath
                break

    # Get path specific for this app
    if appname:
        if path == userDir:
            appname = '.' + appname.lstrip('.')  # Make it a hidden directory
        path = op.join(path, appname)
        if not op.isdir(path):  # pragma: no cover
            os.mkdir(path)

    # Done
    return path


APPDATA_DIR = appdata_dir('webruntime')
RUNTIME_DIR = op.join(APPDATA_DIR, 'webruntimes')
TEMP_APP_DIR = op.join(APPDATA_DIR, 'temp_apps')
DELETE_PREFIX = 'todelete~'


# maybe a bit overkill, but hey, it works!
def versionstring(version):
    """ Given a version string or tuple, produce a version string that looks
    a bit funny but can be string-compared to order versions. Works with
    semver. The only restriction is that any part (between two dots) may not
    be more than 9 chars long.
    """
    if isinstance(version, (tuple, list)):
        version = '.'.join(version)
    if not isinstance(version, str):
        raise TypeError('Version must be a tuple or string.')
    version = version.strip().lower()
    version = version.replace(' ', '').replace('\t', '').replace('~', '')
    
    if version == 'latest':
        return '~~'
    
    isnumeric = True
    anchor = 0
    parts = []
    
    def add_part(i):
        part = version[anchor:i]
        if len(part) > 9:
            raise ValueError('Version parts can be at most 9 chars in %r' % version)
        elif part.isnumeric():
            parts.append('~' + part.rjust(9, ' '))
        elif part:
            parts.append(' ' + part.rjust(9, ' '))
    
    for i in range(len(version)):
        c = version[i]
        if i == anchor:
            isnumeric = c.isnumeric()
        if c == '.':
            add_part(i)
            anchor = i + 1
        elif isnumeric and not c.isnumeric():
            add_part(i)
            anchor = i
            isnumeric = False
    
    add_part(len(version))
    return '.'.join(parts) + '.~'


def get_pid_list():
    """ Get list of pids (on Windows, OS X and Linux)
    """
    if sys.platform.startswith('win'):
        cmd = ['tasklist']
    else:  # Posix
        cmd = ['ps', 'aux']  # Not "-u root -N" cause -N does not work on OS X
    
    out = subprocess.check_output(cmd).decode(errors='ignore')
    pids = []
    for line in out.splitlines():
        parts = [i.strip() for i in line.replace('\t', ' ').split(' ') if i]
        if len(parts) >= 2 and parts[0] != 'root':
            try:
                pids.append(int(parts[1]))
            except ValueError:
                pass
    pids.sort()  # helps in debugging
    if not pids:
        logger.warn('get_pid_list() found zero pids.')
    return pids


def remove(path1, nowarn=False):
    """ Mark a file or directory for removal (by renaming it) and try
    to remove it. If it fails, that's ok; we'll remove it later.
    """
    if not op.exists(path1):
        return
    dirname, name = op.split(path1)
    
    # Rename if we must/can
    if name.startswith(DELETE_PREFIX):
        path2 = path1
    else:
        for i in range(100):
            path2 = op.join(dirname, DELETE_PREFIX + str(i) + name)
            if not op.exists(path2):
                break
        try:
            os.rename(path1, path2)
        except Exception:
            if not nowarn:
                logger.warn('could not clean up %r' % path1)
            return  # if we cannot rename, we cannot delete
    
    # Delete if we can
    try:
        if op.isfile(path2):
            os.remove(path2)
        elif op.isdir(path2):
            shutil.rmtree(path2)
    except Exception:
        pass


def init_dirs():
    """ Ensure that directories exist.
    """
    if not op.isdir(APPDATA_DIR):
        os.mkdir(APPDATA_DIR)
    if not op.isdir(RUNTIME_DIR):
        os.mkdir(RUNTIME_DIR)
    if not op.isdir(TEMP_APP_DIR):
        os.mkdir(TEMP_APP_DIR)


def clean_dirs():
    """ Clean up the webruntime dir and temp app dir.
    """
    
    init_dirs()
    
    # Collect dirs/files and lockfiles
    items = []
    lockfiles = []
    for dir in (RUNTIME_DIR, TEMP_APP_DIR):
        for name in os.listdir(dir):
            path = op.join(dir, name)
            items.append((name, path))
            
            if dir == RUNTIME_DIR and '_' in path and op.isdir(path):
                for lockfilename in os.listdir(path):
                    if lockfilename.startswith('lock~'):
                        try:
                            pid = int(lockfilename.split('~')[-1])
                        except ValueError:
                            continue
                        lockfiles.append((pid, path, lockfilename))
    
    # Get pids, after collecting items, so that we don't remove items for pids
    # that instantiate after getting our list of pids.
    pids = get_pid_list()
    
    # Remove files/dirs that are marked for deletion or associated with a pid
    # that does not exist.
    for name, path in items:
        if '~' in name:
            if name.startswith(DELETE_PREFIX):
                remove(path)
            else:
                try:
                    pid = int(name.split('~')[-1])
                except ValueError:
                    continue  # we probably did not make this
                if pid not in pids:
                    remove(path, True)
    
    # Remove lockfiles in runtime dirs
    dirs_with_lockfiles = set()
    for pid, dir, fname in lockfiles:
        if pid not in pids:
            try:
                os.remove(op.join(dir, fname))
                continue  # i.e. dont mark dir
            except Exception:
                pass
        dirs_with_lockfiles.add(dir)
    
    # Removing old runtimes that are not used (i.e. have no lock files).
    # In theory, a new process could have spawned since we collected the
    # lock files, but because we always use the latest version of a runtime,
    # this should not be a problem.
    runtimes = {}
    for dname in os.listdir(RUNTIME_DIR):
        name, _, version = dname.partition('_')
        if version:
            runtimes.setdefault(name, []).append(version)
    for name in runtimes:
        versions = runtimes[name]
        versions.sort(key=versionstring)
        for version in versions[:-1]:
            dir = op.join(RUNTIME_DIR, name + '_' + version)
            if dir not in dirs_with_lockfiles:
                remove(dir, True)
        # Hack to clear old firefox runtime (it was renamed)
        # remove / deprecate this in version 0.7 or so
        if name == 'xul':
            for version in versions:
                dir = op.join(RUNTIME_DIR, name + '_' + version)
                if dir not in dirs_with_lockfiles:
                    remove(dir, True)


def lock_runtime_dir(path):
    """ Lock a runtime dir for this process.
    """
    assert path.startswith(RUNTIME_DIR)
    lockfile = op.join(path, 'lock~%i' % os.getpid())
    if not op.isfile(lockfile):
        with open(lockfile, 'wb'):
            pass


_app_count = 0

def create_temp_app_dir(prefix, cleanup=60):
    """ Create a temporary direrctory and return its path.

    The directory will be named "<prefix>_<timestamp>_<pid>_<suffix>".
    """
    global _app_count
    _app_count += 1
    
    prefix = prefix.strip(' _-')
    name = '%s_%i_%i~%i' % (prefix, time.time(), _app_count, os.getpid())
    path = op.join(TEMP_APP_DIR, name)
    os.mkdir(path)
    return path


def open_arch(filename):
    """ Open archive, returning the zipfile or tarfile object.
    """
    if filename.endswith(('.tar', '.tar.gz', '.tar.bz2')):
        arch_func = tarfile.open
    elif filename.endswith('.zip'):
        arch_func = zipfile.ZipFile
    return arch_func(filename, mode='r')


def extract_arch(archive, dir_name):
    """ Extract an archive that contains a runtime.
    """
    temp_dir_name = dir_name + '~' + str(os.getpid())
    
    # Maybe the dir is already there ...
    if op.isdir(dir_name):
        return
    
    # Extract it
    t0 = time.time()
    archive.extractall(temp_dir_name)
    
    # Pop out empty dirs
    while True:
        subdirs = os.listdir(temp_dir_name)
        if len(subdirs) != 1 or op.isfile(subdirs[0]) or '.app' in subdirs[0]:
            break
        else:
            pop_dir = subdirs[0] + '~temp'
            os.rename(op.join(temp_dir_name, subdirs[0]),
                    op.join(temp_dir_name, pop_dir))
            for name in os.listdir(op.join(temp_dir_name, pop_dir)):
                os.rename(op.join(temp_dir_name, pop_dir, name),
                        op.join(temp_dir_name, name))
            os.rmdir(op.join(temp_dir_name, pop_dir))
    
    logger.info('Extracted archive into %s in %1.1f s' %
                (op.basename(dir_name), (time.time() - t0)))
    
    # Enable executables for OS X
    for dirpath, dirnames, filenames in os.walk(temp_dir_name):
        if dirpath.endswith('/Contents/MacOS'):
            for fname in filenames:
                if '.' not in fname:
                    filename = op.join(dirpath, fname)
                    os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)
    
    # Mark dir as done
    if op.isdir(dir_name):
        pass  # another process beat us to it
    else:
        os.rename(temp_dir_name, dir_name)
    return True
